"""
This module contains the implementations of the Job concept from a
Gateway configuration.
"""
# ====================== Imports ===============================
from dataclasses import dataclass
import dataclasses
from txp.common.configuration import SamplingWindow
from typing import List, Dict, Union
import datetime
from txp.common.config import settings


# ==============================================================================
# SamplingParameters class
# ==============================================================================
@dataclass
class SamplingParameters:
    """This class encapsulates the sampling parameters that can be assigned to a
    Job or a Job Task

    Note on the args `predicates`: Imagine that we want to cover cases where the sampling
        of a machine depends on something else that is also being monitored. A predicate
        represents a dependency of this kind, with functions that returns a boolean and has
        an ID.

    Args:
        sampling_window: The Sampling Window defined as parameter.

        start_date: Start Date of the Job/Job Task in isoformat YYYY-MM-DD

        end_date: End date of the Job/Job Task in isoformat YYYY-MM-DD. After this date is reached,
            the Job/Job Task will not be executed anymore.

        active_week_days: a week days mask from Sunday to Saturday.
            Example of a mask active from monday to friday: '0111110'

        start_time: Start time of the Job/Job Task within a 24 hrs range, in isoformat HH:MM.

        end_time: Explicit end time of the Job/Job Task within a 24 hrs range, in isoformt HH:MM.

        predicates: A List of predicate names, on which the sampling depends on. For example: 'power_on'.

        samplings: A positive number to indicate how many sampling windows the Driver should execute.
            TODO: This samplings arg should be deleted. Is only provided given the current Driver implementation.
    """
    sampling_window: SamplingWindow
    start_date: datetime.date
    end_date: datetime.date
    active_week_days: str
    start_time: datetime.time
    end_time: datetime.time
    predicates: List[str]

    def __post_init__(self):
        self._week_days_mask: List[int] = list(
            map(lambda char: int(char), self.active_week_days)
        )

    @property
    def week_days_mask(self) -> List[int]:
        """Returns the mask of active week days"""
        return self._week_days_mask

    @classmethod
    def build_from_dict(cls, d: Dict) -> Union["SamplingParameters", None]:
        if not cls._validate_dict(d):
            return None

        sampling_window = SamplingWindow(d['sampling_window']['observation_time'], d['sampling_window']['sampling_time'])
        start_date = datetime.datetime.strptime(d['start_date'], settings.gateway.tasks_scheduler.date_format)
        end_date = datetime.datetime.strptime(d['end_date'], settings.gateway.tasks_scheduler.date_format)
        start_time = datetime.datetime.strptime(d['start_time'], settings.gateway.tasks_scheduler.time_format)
        end_time = datetime.datetime.strptime(d['end_time'], settings.gateway.tasks_scheduler.time_format)

        return cls(
            sampling_window, start_date.date(), end_date.date(), d['active_week_days'], start_time.time(), end_time.time(),
            d['predicates']
        )

    @classmethod
    def _validate_dict(cls, d: Dict) -> bool:
        known_fields = list(map(
            lambda field: field.name in d,
            dataclasses.fields(cls)
        ))

        return all(known_fields)

    def reprJSON(self) -> Dict:
        """Returns a Dict easily serializable to JSON.

        The Dictionary will contain the values required to re-create the instance from the JSON
        object in a pythonic way:
            object = MyObject( **json.loads(json_string) )
        """
        return dict(
            sampling_window=self.sampling_window,
            start_date=self.start_date.strftime(settings.gateway.tasks_scheduler.date_format),
            end_date=self.end_date.strftime(settings.gateway.tasks_scheduler.date_format),
            active_week_days=self.active_week_days,
            start_time=self.start_time.strftime(settings.gateway.tasks_scheduler.time_format),
            end_time=self.end_time.strftime(settings.gateway.tasks_scheduler.time_format),
            predicates=self.predicates
        )


# ==============================================================================
# JobTask class
# ==============================================================================
@dataclass
class JobTask:
    """The encapsulation of a Job Task that is defined inside a Gateway Job.

    Args:
        machines (List[str]): The List of Machines to collect data from during the task.
        parameters (SamplingParameters): The configured parameters for the Job Task.
    """
    machines: List[str]
    parameters: SamplingParameters

    @classmethod
    def build_from_dict(cls, d: Dict) -> Union["JobTask", None]:
        if not cls._validate_dict(d):
            return None

        sampling_parameters = SamplingParameters.build_from_dict(d['parameters'])
        if not sampling_parameters:
            return None

        return cls(d['machines'], sampling_parameters)

    @classmethod
    def _validate_dict(cls, d: Dict) -> bool:
        known_fields = list(map(
            lambda field: field.name in d,
            dataclasses.fields(cls)
        ))

        return all(known_fields)

    def reprJSON(self) -> Dict:
        """Returns a Dict easily serializable to JSON.

        The Dictionary will contain the values required to re-create the instance from the JSON
        object in a pythonic way:
            object = MyObject( **json.loads(json_string) )
        """
        return dict(
            machines=self.machines,
            parameters=self.parameters.reprJSON()
        )


# ==============================================================================
# JobConfig class
# ==============================================================================
@dataclass
class JobConfig:
    """The encapsulation of a Job assigned to the Gateway.

    A Job will define all the sampling operations of the Gateway and the Telemetry dispatch
    operation associated to the monitoring process.

    Args:
        parameters (SamplingParameters): The configured parameters for the job.

        telemetry_rollover_policy (str): The selected policy for the Telemetry dispatching to follow
            when no all packages produced during an sampling window were sent in that window.

        Tasks (List[JobTask]): The List of Job Tasks configured for the monitoring.
    """
    parameters: SamplingParameters
    telemetry_rollover_policy: str
    tasks: List[JobTask]

    @classmethod
    def build_from_dict(cls, d: Dict) -> Union["JobConfig", None]:
        if not cls._validate_dict(d):
            return None

        job_params = SamplingParameters.build_from_dict(d['parameters'])

        if not job_params:
            return None

        tasks = list(map(
            lambda task_dict: JobTask.build_from_dict(task_dict),
            d['tasks']
        ))

        if any(map(
            lambda t: t is None, tasks
        )):
            return None

        return cls(job_params, d['telemetry_rollover_policy'], tasks)

    @classmethod
    def _validate_dict(cls, d: Dict) -> bool:
        known_fields = list(map(
            lambda field: field.name in d,
            dataclasses.fields(cls)
        ))

        return all(known_fields)

    def reprJSON(self) -> Dict:
        """Returns a Dict easily serializable to JSON.

        The Dictionary will contain the values required to re-create the instance from the JSON
        object in a pythonic way:
            object = MyObject( **json.loads(json_string) )
        """
        return dict(
            parameters=self.parameters,
            telemetry_rollover_policy=self.telemetry_rollover_policy,
            tasks=list(map(
                lambda task: task.reprJSON(),
                self.tasks
            )),
        )



"""
This module implements the TasksScheduler object which provides
timer capabilities to other components.
"""
# ============================ imports =========================================
from typing import Optional, List, Callable, Tuple, Dict
from dataclasses import dataclass
from txp.common.config import settings
from txp.common.configuration import SamplingWindow
from txp.common.configuration.job import JobConfig, JobTask
from txp.devices.devices_manager import DevicesManager
import enum
import datetime
import threading
import time
import schedule
import logging
import copy

log = logging.getLogger(__name__)

log.setLevel(settings.txp.general_log_level)


@dataclass
class ScheduleTask:
    """The encapsulation of an Scheduled Task that can be ingested and executed by the TasksScheduler.

    It contains all the information required to configure the schedule task in time.

    And also requires from client classes to configure a callable that will be executed
    as "action" of the schedule time when the time comes. This allows for the TaskScheduler to deal
    only with time handling, while decoupling the actions performed over time as a responsibility
    of the client classes.

    """

    start_date: datetime.date  # Start time date with the format "%Y-%m-%d %H:%M:%S"
    end_date: datetime.date
    start_time: datetime.time
    end_time: datetime.time
    active_week_days: List
    callback: Callable
    callback_args: Tuple = ()


class TasksScheduler:
    """The TasksScheduler object provides an API for client classes to schedule
    tasks in time.

    A ScheduleTask contains the following attributes that will be interpreted accordingly:
        - A Start Date, which represents the specific Date and Time when a task will be
            executed for the first time.
            The Start Date format is specified in the configuration: "%Y-%m-%d %H:%M:%S"

        - Frequency: A Frequency enumerated value which indicates the frequency on which the task
        will be repeated.

        - Callback: A Callable name (method, function) provided by the client class to be the action executed
            in the scheduled task.

        - Callback Arguments: A tuple of positional arguments to forward to the Callback.
    """

    def __init__(self, name: str):
        """
        Args:
            name: Set a name to the TaskScheduler instance. This is useful for logging.
            tasks: The List of Tasks to be scheduled.
        """
        self._name = name
        self._scheduler_run_thread: Optional[threading.Thread] = None
        self._stop_event: threading.Event = None
        self._scheduler: schedule.Scheduler = schedule.Scheduler()

    def schedule_daily_callback(self, callback, start_time: str) -> None:
        """Schedule an arbitrary callback to be executed daily at start time.
        """
        log.info(f"Scheduler set daily scheduled callback to run at: {start_time}")
        self._scheduler.every().day.at(start_time).do(callback)

    def schedule_task_once(self, task: ScheduleTask, daily: bool = True, log_info: bool = True) -> None:
        """Schedule the tasks in the schedule.Scheduler instance
        to run just only once.

        The task callback is wrapped around a closure which returns
        the schedule.CancelJob value. This is done to execute the
        scheduled task only once.

        https://schedule.readthedocs.io/en/stable/examples.html#run-a-job-once
        """

        def run_just_once():
            task.callback(*task.callback_args)
            return schedule.CancelJob

        desired_daily_time_str = task.start_time.strftime("%H:%M:%S")

        if daily:
            self._scheduler.every().day.at(desired_daily_time_str).do(run_just_once)
        else:
            self._scheduler.every().second.do(run_just_once)

        scheduled_tasks = self._scheduler.jobs
        if scheduled_tasks and log_info:
            scheduled_task_datetime = scheduled_tasks[-1].next_run

            log.info(
                f"{self._name} new task to run "
                f"the {scheduled_task_datetime.date()} at {scheduled_task_datetime.time()} hours"
            )

    def is_schedule_required(self, task: ScheduleTask) -> bool:
        """Returns True if the task requires to be schedule give the current day.

        Currently only validates if the Date is valid.
        """
        # Validate the ongoing date
        current_datetime = datetime.datetime.now()
        current_date = datetime.datetime.now().date()
        valid_date = (
            (task.start_date == current_date)
            or (current_date < task.end_date)
            or (current_date == task.end_date)
        )

        # Validate the week day
        current_week_day = datetime.datetime.now().weekday()
        valid_week_day = bool(task.active_week_days[current_week_day])

        # Validate the hour
        task_datetime = datetime.datetime.combine(
             task.start_date, task.start_time
        )

        if task_datetime.date() > current_datetime.date():  # Potentially, next day.
            valid_time = True
        else:
            valid_time = task_datetime.time() >= current_datetime.time()

        return valid_week_day and valid_date and valid_time

    def is_task_currently_active(self, task: ScheduleTask) -> bool:
        """Returns True if the task requires to be schedule and started now,
        because it's currently active given the configuration."""
        current_date = datetime.datetime.now().date()
        valid_date = (
            (task.start_date == current_date)
            or (current_date < task.end_date)
            or (current_date == task.end_date)
        )

        now = datetime.datetime.now()
        valid_hour = (
            task.start_time < now.time() < task.end_time
        )

        # Validate the week day
        current_week_day = datetime.datetime.now().weekday()
        valid_week_day = bool(task.active_week_days[current_week_day])

        return valid_week_day and valid_date and valid_hour

    def _run_schedule(self):
        log_counter = 0
        while True:
            if self._stop_event.is_set():
                log.info("Stopping scheduled tasks")
                self._scheduler.clear()
                break

            self._scheduler.run_pending()
            next_run = self._scheduler.next_run
            if (
                next_run
                and (next_run - datetime.datetime.now()).total_seconds() > 0
                and log_counter > 100
            ):
                log.info(
                    f"{self._name} task scheduler next run will be at: {self._scheduler.next_run}"
                )
                log_counter = 0

            elif next_run is None and log_counter > 100:
                log.info(
                    "{self._name} task scheduler has no more scheduled tasks pending"
                )

            log_counter = log_counter + 1
            time.sleep(settings.gateway.tasks_scheduler.check_stop_event_interval)

    def start(self) -> None:
        """Starts the execution of the tasks contained in the scheduler in
        a separate thread.

        This call will return immediately.
        """

        if (
            self._scheduler_run_thread is not None
            and self._scheduler_run_thread.is_alive()
        ):
            raise RuntimeError("The scheduled already has been started")

        self._stop_event = threading.Event()
        self._scheduler_run_thread = threading.Thread(
            target=self._run_schedule,
            name="scheduler_run_thread",
            daemon=True,
        )

        self._scheduler_run_thread.start()
        log.debug("TaskScheduler scheduling thread started")

    def stop(self) -> None:
        """Stops the scheduler execution of tasks.

        It gives guarantee to stop the execution of future tasks after
        the current one has been executed.
        """
        if (
            self._scheduler_run_thread is None
            or not self._scheduler_run_thread.is_alive()
        ):
            raise RuntimeError(
                "Calling to Stop scheduler, but no scheduler as been started."
            )

        self._stop_event.set()
        log.info("Requesting to stop scheduling thread")

    def is_running_tasks(self) -> bool:
        """Returns True if the scheduler is currently busy with the processing
        of scheduling tasks. That means, the self._scheduler_run_thread is actively running"""
        return self._scheduler is not None and self._scheduler_run_thread.is_alive()

    def get_scheduled_tasks_dict(self) -> Dict:
        """Return a dictionary of the scheduled tasks to be executed.

        Returns:
            A Dictionary of scheduled tasks to be executed. The values are
            datetime values.
        """
        jobs = self._scheduler.jobs
        jobs = dict(map(lambda tpl: (tpl[0], tpl[1].next_run), enumerate(jobs)))
        return jobs



from typing import List
from txp.ml.common.tasks.slic_cv_tasks import SlicPatchRecognitionTask
from txp.ml.prediction_service.state_managers.axionlog import RayBasicAxionlogGateAsset
from txp.ml.prediction_service.state_managers.gao import RayBasicGaoAsset
from txp.ml.prediction_service.state_managers.beer_showrrom_states import RayBasicShowroomBeerAsset
from txp.ml.prediction_service.state_managers.showroom import RayBasicShowroomVibrationAsset
from txp.ml.prediction_service.state_managers.mock_state_manager import RayMockAsset
from txp.ml.common.tasks.vibration_basic_task import VibrationBasicTask
from txp.ml.common.tasks import MockTask
from txp.common.ml.tasks import AssetTask
import logging


class TaskFactory:
    """
        Tasks factory used for creating all ml task objects.
    """
    @staticmethod
    def create_task(task_definition: AssetTask, credentials_str):
        # TODO: this cases of task_type should come from firestore

        task_type = task_definition.task_type
        if task_type == 'VibrationBasicTask':
            logging.info(f"Creating task for VibrationBasicTask")
            return VibrationBasicTask(task_definition, credentials_str)
        elif task_type == 'VibrationBasicGradientBoostTask':
            logging.info(f"Creating task for VibrationBasicGradientBoostTask")
            return VibrationBasicTask(task_definition, credentials_str, decision_tree_policy=False)
        elif task_type == 'SlicPatchRecognitionTask':
            logging.info(f"Creating task for SlicPatchRecognitionTask")
            return SlicPatchRecognitionTask(task_definition, credentials_str)
        elif task_type == 'MockTask':
            return MockTask(task_definition, credentials_str)
        else:
            logging.info(f"Task type not supported for task: {task_definition.dict()}")

    @staticmethod
    def create_tasks_for_asset(asset_id, assets_config, credentials_str):
        tasks = {}
        for task in assets_config[asset_id]['tasks'].values():
            task_definition = AssetTask(**task)
            tasks[task['task_id']] = TaskFactory.create_task(task_definition, credentials_str)
        return tasks

    @staticmethod
    def create_tasks(assets_config, state_managers, credentials_str):
        tasks = {}
        for asset_id in assets_config:
            if assets_config[asset_id]["state_manager"] not in state_managers:
                continue
            tasks = {**tasks, **TaskFactory.create_tasks_for_asset(asset_id, assets_config, credentials_str)}
        for task in tasks.values():
            if not task.is_ready():
                logging.info(f"Error starting task for task_id: {task.task_id}")
            else:
                logging.info(f"Task {task.task_id} ready")
        return tasks


class StateManagerFactory:
    """
        Tasks factory used for creating all ml state manager objects.
    """

    @staticmethod
    def create_state_managers(assets_config, credentials_str, backup_collection_name, events_and_states_dataset,
                              notifications_topic):
        state_managers = {}
        for asset_id in assets_config:
            state_m = StateManagerFactory.create_state_manager_for_asset(asset_id, assets_config[asset_id],
                                                                         credentials_str, backup_collection_name,
                                                                         events_and_states_dataset, notifications_topic)
            if state_m is not None:
                state_managers[assets_config[asset_id]["state_manager"]] = state_m

        return state_managers

    @staticmethod
    def create_state_manager_for_asset(asset_id, asset_config, credentials_str, backup_collection_name,
                                       events_and_states_dataset, notifications_topic):
        state_manager_id = asset_config["state_manager"]
        tasks = []
        for task in asset_config["tasks"].values():
            tasks.append(task["task_id"])

        if "motor_lab_sm" == state_manager_id:
            return RayBasicShowroomVibrationAsset.remote(asset_id, tasks, events_and_states_dataset,
                                                         backup_collection_name, notifications_topic, credentials_str)
        elif "beer_lab_sm" == state_manager_id:
            return RayBasicShowroomBeerAsset.remote(asset_id, tasks, events_and_states_dataset, backup_collection_name,
                                                    notifications_topic, credentials_str)
        elif "gao_sm" == state_manager_id:
            asset_tasks: List[AssetTask] = list(map(
                lambda task_dict: AssetTask(**task_dict),
                asset_config["tasks"].values()
            ))
            return RayBasicGaoAsset.remote(asset_id, tasks, events_and_states_dataset, backup_collection_name,
                                           asset_tasks, notifications_topic, credentials_str)
        elif "axionlog_sm" == state_manager_id:
            asset_tasks: List[AssetTask] = list(map(
                lambda task_dict: AssetTask(**task_dict),
                asset_config["tasks"].values()
            ))
            return RayBasicAxionlogGateAsset.remote(asset_id, tasks, events_and_states_dataset, backup_collection_name,
                                                    asset_tasks, notifications_topic, credentials_str)

        elif "mock_sm" == state_manager_id:
            return RayMockAsset.remote(asset_id, tasks, events_and_states_dataset, backup_collection_name,
                                       notifications_topic, credentials_str)
        else:
            logging.info(f"state manager not supported, state_manager_id: {state_manager_id}, asset_id: {asset_id}")
            return None



import json
from google.oauth2 import service_account
from google.cloud import bigquery
import ray
from txp.ml.prediction_service.service_manager.actors_synchronizer import stateful_method
from google.cloud import pubsub_v1
import logging


class StateManagerOutput:
    """ This class is in charge of managing the output for a given state manager.
    A notification is going to be sent each time a new state is stored in bigquery.
    Args:
        events_and_states_dataset: BQ dataset name where events and state are going to be stored.
        notifications_topic: PubSub topic id where notifications are going to be sent.
        credentials_str: GCP credentials string.
    """

    def __init__(self, events_and_states_dataset, notifications_topic, credentials_str, log_level=logging.INFO):
        logging.basicConfig(level=log_level)
        self.events_table_id = "events"
        self.states_table_id = "states"
        self.notifications_topic = notifications_topic
        json_dict_service_account = json.loads(credentials_str, strict=False)
        self.credentials = service_account.Credentials.from_service_account_info(json_dict_service_account)
        self.bigquery_client = bigquery.Client(credentials=self.credentials, project=self.credentials.project_id)
        self.dataset = f"{self.credentials.project_id}.{events_and_states_dataset}"

    def __insert_rows_to_bigquery(self, rows, table_id):
        errors = self.bigquery_client.insert_rows_json(f"{self.dataset}.{table_id}", rows)
        if errors:
            logging.info(f"""Could not store in {table_id}: {rows} """)
            return False
        else:
            return True

    @stateful_method
    def insert_events_to_bigquery(self, events):
        res = self.__insert_rows_to_bigquery(events, self.events_table_id)
        if res:
            logging.info(f"{len(events)}, Events were stored in bigquery")

    @stateful_method
    def insert_state_to_bigquery(self, state):
        res = self.__insert_rows_to_bigquery([state], self.states_table_id)
        if res:
            logging.info(f'new state stored to bigquery: {state["observation_timestamp"]}')

    @stateful_method
    def publish_to_pubsub(self, state):
        publisher = pubsub_v1.PublisherClient(credentials=self.credentials)
        topic_path = publisher.topic_path(self.credentials.project_id, self.notifications_topic)
        future = publisher.publish(topic_path, data=json.dumps(state).encode('utf-8'))
        future.result()
        logging.info(f"New State published to PubSub {state}")


@ray.remote
class RayStateManagerOutput(StateManagerOutput):
    def __init__(self, events_and_states_dataset, notifications_topic, credentials_str):
        super().__init__(events_and_states_dataset, notifications_topic, credentials_str)



"""
This module implements the MainView for the layout of the application.
The main view is composed of the view area and the submenu on the right.
"""
from txp.web.core_components.app_component import AppComponent, AppState
from txp.web.core_components.app_profiles import AppProfile
import abc
from typing import Dict
import streamlit as st


class MainView(AppComponent):
    """The MainView component for the application.

    A MainView represents the View for one selected panel.
    For example, some views are:
        - Under 'Resumen' profile, the 'Telemetry Health' view.
        - Under 'Resumen' profile, the 'Devices Health' view.

    Each MainView is composed of two columns:
        1) The content area column: Where the content is actually displayed. It's the wider section on screen.
        2) The submenu area column: The submenu shown on the Right hand side inside a smaller column.

    The _render method is a provided template which renders both columns on each streamlit rerun.
    Concrete implementations should be provided by child classes.
    """
    def __init__(self, component_key: str):
        super(MainView, self).__init__(
            component_key=component_key
        )
        self._submenu_state: Dict = self._build_submenu_state()
        self._content_state: Dict = self._build_content_state()

    @classmethod
    def _register_app_state(cls, state: AppState) -> None:
        raise NotImplemented

    @abc.abstractmethod
    def get_associated_profile(self) -> AppProfile:
        pass

    @abc.abstractmethod
    def _build_submenu_state(self) -> Dict:
        pass

    @abc.abstractmethod
    def _build_content_state(self) -> Dict:
        pass

    @abc.abstractmethod
    def _render_content(self) -> None:
        pass

    @abc.abstractmethod
    def _render_submenu(self) -> None:
        pass

    @classmethod
    @abc.abstractmethod
    def get_display_name(cls) -> str:
        """Returns the human readable name for this view, to be shown in the Menu of the application."""
        pass

    def _render(self) -> None:
        main_view_col, submenu_col = st.columns([.75, .25])
        with main_view_col:
            self._render_content()
        with submenu_col:
            self._render_submenu()



from txp.web.core_components.main_view import MainView
from txp.web.core_components.app_profiles import AppProfile
import logging
from typing import Dict
import streamlit as st
import pandas as pd
import numpy as np
import plotly.graph_objects as go
import pytz
from txp.web.core_components.app_component import Card
from txp.common.config import settings
import txp.cloud.controllers.bigquery as bigquery_controllers
import txp.cloud.controllers.firestore as firestore_controller
import datetime

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


class AssetConditionsView(MainView):
    """The Device Health View"""

    def __init__(self):
        super(AssetConditionsView, self).__init__(component_key="asset_conditions_view")
        self.current_machine = None
        self.current_selected_assets = None
        self.current_machines_groups = None
        self.data = None
        self.report_type = None
        self.events = {}
        self.states = {}
        self.states_proportion = {}
        self._global_report = "Reporte global"
        self._specific_report = "Reporte por activo"
        self._default_value_report = f"Sin registro en los últimos {self._days_report} días"
        self.now = None
        self._monitoring_start = None
        self.color_map = {
            'Óptima': '3D62F6',
            'Buena': '79CF24',
            'Operativa': 'F6F03D',
            'Indefinida': '808080',
            'Crítica': 'F64F3D'
        }
        self.formatted_asset_name = None

        # The control attributes defined below are required for optimization purposes
        #   across streamlit reruns
        self._redraw_required: bool = True
        self._assets_resume_table: Card = None
        self._conditions_poltly_fig: go.Figure = None

    def get_associated_profile(self) -> AppProfile:
        return AppProfile.RESUMEN

    @classmethod
    def get_display_name(cls) -> str:
        return "Condición de los Activos"

    def _build_submenu_state(self) -> Dict:
        return {}

    def _build_content_state(self) -> Dict:
        return {}

    def _formatted_data(self, data):
        """
        takes the project data and returns a formatted Dict.
        """
        machines_groups = {d["name"]: d for d in data["machines_groups"]}
        machines = {d["machine_id"]: d for d in data["machines"]}
        gateways = {d["gateway_id"]: d for d in data["gateways"]}
        edges = {d["logical_id"]: d for d in data["edges"]}

        data = {
            "machines_groups": machines_groups,
            "machines": machines,
            "gateways": gateways,
            "edges": edges,
        }
        return data

    def update_button_clicked(self):
        self._redraw_required = True
        self.get_data()

    def get_data(self):
        if self.data is None:
            db = firestore_controller.get_authenticated_client(
                self.app_state.authentication_info.role_service_account_credentials
            )

            data = firestore_controller.pull_recent_project_data(db, self._tenant_id)
            self.data = self._formatted_data(data)
            machines = list(self.data["machines"].keys())
            self.current_selected_assets = machines
            machines_groups = list(self.data["machines_groups"].keys())
            self.current_machines_groups = machines_groups[0]
            self.current_machine = machines[0]
            self.formatted_asset_name = self.formatting_readable_strings(self.current_machine)
            db.close()

            self.bqclient = bigquery_controllers.get_authenticated_client(
                self.app_state.authentication_info.role_service_account_credentials
            )

        self._monitoring_start, self.now = self.monitoring_range()
        log.info(f"Your intervals for asset conditions: {self._monitoring_start} - {self.now}")

        # TODO: We can improve this with only 1 query !!!!
        #   And then just 'groupby' the received dataset

        for asset in self.data["machines"].keys():
            to_states_distribution = bigquery_controllers.get_all_task_predictions_within_interval(
                    self._tenant_id,
                    self._states_dataset,
                    asset,
                    self._monitoring_start.astimezone(pytz.utc),
                    self.now.astimezone(pytz.utc),
                    self.bqclient
                )
            self.states_proportion.update({asset: to_states_distribution})

        for asset in self.data["machines"].keys():
            last_asset_state = bigquery_controllers.get_last_task_prediction_for_asset(
                self._tenant_id,
                self._states_dataset,
                asset,
                self._days_report,
                self.bqclient
            )

            self.states.update({asset: last_asset_state})

    def monitoring_range(self):
        timezone = pytz.timezone("America/Mexico_City")
        mex_now = datetime.datetime.now(tz=timezone)
        mex_now_midnight = datetime.datetime.combine(
            mex_now.date(),
            datetime.time(hour=0, minute=0, second=0, microsecond=0)
        )
        mex_now_midnight = timezone.localize(mex_now_midnight)
        return mex_now_midnight, mex_now

    def change_utc_timezone(self, timestamp, date=True):
        timezone = pytz.timezone("America/Mexico_City")
        date_time = pd.to_datetime(timestamp, utc=True)
        new_timezone = date_time.astimezone(timezone)
        if date is False:
            strtime = new_timezone.strftime("%H:%M:%S")
        else:
            strtime = new_timezone.strftime("%m/%d/%Y %H:%M:%S")
        return strtime

    def utc_timestamp_to_local_datetime(self, timestamp):
        timezone = pytz.timezone("America/Mexico_City")
        date_time = pd.to_datetime(timestamp, utc=True)
        new_timezone = date_time.astimezone(timezone)
        return new_timezone

    def _render_submenu(self) -> None:
        st.markdown('\n')
        st.markdown('\n')
        assets = self.data['machines'].keys()
        st.markdown("**<p align='center'>Estado general</p>**", unsafe_allow_html=True)
        st.markdown('\n')
        st.markdown(f"**Activos monitoreados**: {len(assets)}")

        self.download_data_info_message()

        m = st.markdown("""
        <style>
        div.stButton > button:first-child {
        color: #0000CD;
        box-sizing: 5%;
        height: 2em;
        width: 100%;
        font-size:20px;
        border: 2px solid;
        border-color: #D3D3D3;
        border-radius: 10px;
        box-shadow: 1px 1px 1px 1px 
        rgba(0, 0, 0, 0.1);
        padding-top: 0px;
        padding-bottom: 0px;
        }
        .center {
        margin-left: auto;
        margin-right: auto;
        }
        </style>""", unsafe_allow_html=True)

        st.button('Actualizar', on_click=self.get_data)

    def get_last_event_for_state(self, events):
        event_timestamps = [pd.to_datetime(event['observation_timestamp']) for event in events]
        for event in events:
            event_timestamp = pd.to_datetime(event['observation_timestamp'])
            if event_timestamp == max(event_timestamps):
                return event
            else:
                pass

    def resume_table(self):
        if self._redraw_required:
            conditions = []
            dates = []

            for asset, state in self.states.items():
                if not state:
                    condition = self._default_value_report
                    date = self._default_value_report
                else:
                    condition = state['condition']
                    date = self.change_utc_timezone(state['observation_timestamp'])
                conditions.append(condition)
                dates.append(date)

            condition_list = []
            for condition in conditions:
                if not condition == self._default_value_report:
                    condition_list.append(self.app_state.condition_report.conditions_mapping[condition])
                else:
                    condition_list.append(self._default_value_report)

            resumen = {
                "Activo": [self.formatting_readable_strings(f_asset) for f_asset in self.states.keys()],
                "Condition": condition_list,
                "Date": dates,
            }

            self._assets_resume_table = Card(
                card_row=1,
                card_column_start=1,
                card_column_end=4,
                card_subheader="Resumen de los activos",
                content=resumen,
                labels=[
                    "Activo",
                    "Condición",
                    "Fecha de cambio de condición",
                ],
                vertical_style=False,
            )

        self.render_card(
            [
                self._assets_resume_table,
            ]
        )

    def download_data_info_message(self):
        strtime = self.now.strftime("%m/%d/%Y, %H:%M:%S")
        st.caption(f"Última actualización {strtime}")

    def states_proportion_plot(self, asset):
        if not self.states_proportion[asset].empty:
            if self._redraw_required:
                states_counts_serie: pd.Series = self.states_proportion[asset].condition.value_counts()
                states_count_dict = states_counts_serie.to_dict()
                fig = go.Figure(
                    data=[
                        go.Pie(
                            labels=list(states_count_dict.keys()),

                            values=list(states_count_dict.values()),
                            pull=[0.1, 0],
                        )
                    ]
                )
                fig.update_layout(
                    height=350,
                    width=450,
                    margin=dict(l=0, r=0, t=0, b=0),
                    font=dict(size=20)
                )
                generated_values_color_map = {}
                colors = np.array([''] * len(list(states_count_dict.values())), dtype=object)
                for i, (k, v) in enumerate(states_count_dict.items()):
                    generated_values_color_map[v] = self.color_map[self._asset_states_sp[k]]
                    colors[i] = generated_values_color_map[v]

                fig.update_traces(marker=dict(colors=colors))
                fig.update_layout(legend=dict(
                    xanchor="left",
                    yanchor="top",
                    x=-0.15,
                    font=dict(
                        size=16,
                        color="black"
                    ),
                ))
                fig.update_layout(
                    height=350,
                    width=450,
                    margin=dict(l=0, r=0, t=0, b=0),
                    font=dict(size=20))

                self._conditions_poltly_fig = fig

            # Draw the figure
            first_condition = self.states_proportion[asset]._get_value(0, 'observation_timestamp')
            log.info(f"The start raw timestamp is: {first_condition}")
            strtime: datetime.datetime = self.utc_timestamp_to_local_datetime(first_condition)
            log.info(f"The value of your strtime based on your timestamp is: {strtime}")
            last_condition = self.states_proportion[asset].iloc[-1]['observation_timestamp']
            strlasttime = self.utc_timestamp_to_local_datetime(last_condition)
            st.markdown(f"#### Registro de {strtime.strftime('%H:%M:%S')} a {strlasttime.strftime('%H:%M:%S')} del "
                        f"{strlasttime.strftime('%d-%m-%Y')}")
            st.markdown("##### Distribución de las condiciones", unsafe_allow_html=True)
            st.plotly_chart(fig, use_container_width=True)
            st.markdown("***")

        else:
            st.markdown("Sin registro en el último día")

    def _render_content(self) -> None:
        if self.data is None:
            self.get_data()
        e = st.empty()
        with e.container():
            st.markdown(f"## {self._global_report} de la condición de los activos")
            self.resume_table()
            st.markdown(f"## {self._specific_report} de la condición de los activos")
            assets = self.data['machines'].keys()
            formated_assets = [self.formatting_readable_strings(f_asset) for f_asset in assets]
            for formated_asset, asset in list(zip(formated_assets, assets)):
                st.markdown(f"### {formated_asset}")
                self.states_proportion_plot(asset)



import logging
from txp.ml.training_service.dataset_fetchers.bigquery_dataset_fetcher import (
    BigqueryDatasetFetcher,
)
import ray
from txp.cloud.common import utils as cu
from txp.ml.training_service.model_trainers.basic_model_trainer import BasicModelTrainer
from txp.common.ml import models
from txp.ml.training_service import training_exceptions as te


class VibrationModelTrainer(BasicModelTrainer):
    def __init__(
            self,
            credentials_str,
            models_bucket_name,
            data_warehouse_bucket_name,
            prediction_service_topic,
            models_registry_collection_name,
            log_level: logging.INFO,
    ):
        super().__init__(
            credentials_str, models_bucket_name, prediction_service_topic, models_registry_collection_name, log_level
        )
        self.fetcher = BigqueryDatasetFetcher(
            credentials_str, data_warehouse_bucket_name, log_level
        )

    def _train_task(
            self,
            task,
            task_definition,
            dataset_name,
            dataset_versions,
            tenant_id,
            machine_id,
            task_id,
    ):
        logging.basicConfig(level=self.log_level)
        logging.info(
            f"Starting to Download data for {tenant_id}, {machine_id}, {task_id}, {dataset_name}"
        )
        gcs_task_path = cu.get_gcs_task_bucket_path(tenant_id, machine_id, task_id)
        time_df = self.fetcher.get_time_dataframe_from_gcs(
            dataset_name, dataset_versions, gcs_task_path
        )

        time_df["label"].loc[:, "label"] = time_df["label"].apply(lambda x: self.get_label_value(task_definition, x))

        dataframes = {
            "time_df": time_df,
            "fft_df": self.fetcher.get_fft_dataframe_from_gcs(
                dataset_name, dataset_versions, gcs_task_path
            ),
            "psd_df": self.fetcher.get_psd_dataframe_from_gcs(
                dataset_name, dataset_versions, gcs_task_path
            ),
            "time_metrics_df": self.fetcher.get_time_metrics_dataframe_from_gcs(
                dataset_name, dataset_versions, gcs_task_path
            ),
            "fft_metrics_df": self.fetcher.get_fft_metrics_dataframe_from_gcs(
                dataset_name, dataset_versions, gcs_task_path
            ),
            "psd_metrics_df": self.fetcher.get_psd_metrics_dataframe_from_gcs(
                dataset_name, dataset_versions, gcs_task_path
            ),
        }
        logging.info(
            f"Data Downloaded for {tenant_id}, {machine_id}, {task_id}, {dataset_name}"
        )
        logging.info(
            f"Training model for {tenant_id}, {machine_id}, {task_id}, {dataset_name}"
        )
        try:
            dataset = task.encoder.build_training_dataset(dataframes, "label")
            task.policy.train(dataset, "label")
            storage_task = task.save()
        except:
            raise te.ErrorWhileTrainingModel()

        return storage_task

    def get_model_registry(self, task, tenant_id, machine_id, task_id, file_path) -> models.ModelRegistry:
        state = models.ModelState(
            value=models.ModelStateValue.TRAINING,
            creation_date=cu.get_current_utc_datetime_str(),
            publishment_date="",
            deprecation_date=""
        )
        metadata = models.ModelMetadata(
            feedback=models.DefaultModelFeedback(),
            tenant_id=tenant_id,
            machine_id=machine_id,
            task_id=task_id
        )
        registry = models.ModelRegistry(
            metadata=metadata,
            file_path=file_path,
            state=state
        )
        return registry

    def get_model_feedback(self, task) -> models.ModelFeedback:
        return models.SuccessModelFeedback(metrics={"accuracy": task.policy.accuracy})


@ray.remote
class RayVibrationModelTrainer(VibrationModelTrainer):
    def __init__(
            self,
            credentials_str,
            models_bucket_name,
            data_warehouse_bucket_name,
            prediction_service_topic,
            models_registry_collection_name,
            log_level=logging.INFO,
    ):
        super().__init__(
            credentials_str,
            models_bucket_name,
            data_warehouse_bucket_name,
            prediction_service_topic,
            models_registry_collection_name,
            log_level,
        )



import datetime

from txp.common.ml.annotation import AnnotationLabel
from txp.cloud.controllers.authentication import AuthenticationInfo
from txp.web.core_components.app_component import Card
import matplotlib.pyplot as plt
from txp.common.plotter_utils import FFTPlotter, PSDPlotter
from txp.cloud.common import metrics
import txp.cloud.controllers.bigquery as bigquery_controllers
import pytz
import time
import logging
import streamlit as st
import plotly
import plotly.graph_objects as go
import numpy as np
import os
import pandas as pd

plt.rcParams.update({"font.size": 6.5})
current_directory = os.getcwd()

# TODO: This might be a problem. The logging configuration is being taken from `txp` package
from txp.common.config import settings

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


def mapping_data_from_data_parent(parent):
    list_of_element = []
    for element in parent:
        data = element.reprJSON()
        list_of_element.append(data)
    return list_of_element


def get_machines_with_vibration_analysis(edges_data):
    vibrational_assets = {}
    for machine_id, edges in edges_data.items():
        if any(edge["device_kind"] == "Icomox" for edge in edges):
            machine = machine_id
            v_edges = []
            for edge in edges:
                if edge["device_kind"] == "Icomox":
                    v_edges.append(edge["logical_id"])
            vibrational_assets[machine] = v_edges

    return vibrational_assets


def vibrational_perceptions(edges_data, current_machine, current_edge, project_data):
    for edge in edges_data[current_machine]:
        if edge["logical_id"] == current_edge:
            for device in project_data.device_kinds_table.values():
                if edge["device_kind"] == device["kind"]:
                    return list(device["perceptions"].keys())


def add_trace(format_type, timestamps, data, current_perception, labels, user):
    """Adds traces per sampling window data and set visibility buttons."""

    traces = []
    for timestamp, objects, label in zip(timestamps, data, labels):
        if label is not None and user == AuthenticationInfo.UserRole.Admin:
            name = f"{label}: "
            x_value = name + timestamp.strftime("%H:%M:%S")
        else:
            x_value = timestamp.strftime("%H:%M:%S")
        for object, visibility in zip(objects, [True, False, False]):
            if format_type == "fft":
                trace = go.Scatter3d(
                    x=[x_value] * len(object.frequency_axis),
                    y=object.frequency_axis,
                    z=object.sample_count
                    * np.abs(object.signal_frequency_amplitude_axis),
                    mode="lines",
                    name=f"{x_value}",
                    visible=visibility,
                )
                traces.append(trace)
            elif format_type == "psd":
                trace = go.Scatter3d(
                    x=[x_value] * len(object.frequency),
                    y=object.frequency,
                    z=object.psd,
                    mode="lines",
                    name=f"{x_value}",
                    visible=visibility,
                )
                traces.append(trace)

    if not current_perception == "Microphone":
        layout = dict(
            updatemenus=[
                dict(
                    type="buttons",
                    showactive=True,
                    buttons=list(
                        [
                            dict(
                                label="x",
                                method="update",
                                args=[{"visible": [True, False, False] * len(data)}],
                            ),
                            dict(
                                label="y",
                                method="update",
                                args=[{"visible": [False, True, False] * len(data)}],
                            ),
                            dict(
                                label="z",
                                method="update",
                                args=[{"visible": [False, False, True] * len(data)}],
                            ),
                        ]
                    ),
                )
            ],
        )
    else:
        layout = dict(height=650)
    fig = go.Figure(data=traces, layout=layout)
    return fig


def plot_per_condition(format_type, object_data, current_perception):
    fig = go.Figure()
    if format_type == "fft":
        fig.add_trace(
            go.Scatter(
                x=object_data[0].frequency_axis,
                y=object_data[0].sample_count
                * np.abs(object_data[0].signal_frequency_amplitude_axis),
            )
        )

    elif format_type == "psd":
        fig.add_trace(
            go.Scatter(
                x=object_data[0].frequency,
                y=object_data[0].psd,
            )
        )

    if not current_perception == "Microphone":
        fig.update_layout(
            updatemenus=[
                dict(
                    type="buttons",
                    showactive=True,
                    buttons=list(
                        [
                            dict(
                                label="x",
                                method="update",
                                args=[
                                    {
                                        "y": [
                                            object_data[0].sample_count
                                            * np.abs(
                                                object_data[
                                                    0
                                                ].signal_frequency_amplitude_axis
                                            )
                                            if format_type == "fft"
                                            else object_data[0].psd,
                                        ],
                                        "x": [
                                            object_data[0].frequency_axis
                                            if format_type == "fft"
                                            else object_data[0].frequency
                                        ],
                                    }
                                ],
                            ),
                            dict(
                                label="y",
                                method="update",
                                args=[
                                    {
                                        "y": [
                                            object_data[1].sample_count
                                            * np.abs(
                                                object_data[
                                                    1
                                                ].signal_frequency_amplitude_axis
                                            )
                                            if format_type == "fft"
                                            else object_data[1].psd
                                        ],
                                        "x": [
                                            object_data[1].frequency_axis
                                            if format_type == "fft"
                                            else object_data[1].frequency
                                        ],
                                    }
                                ],
                            ),
                            dict(
                                label="z",
                                method="update",
                                args=[
                                    {
                                        "y": [
                                            object_data[2].sample_count
                                            * np.abs(
                                                object_data[
                                                    2
                                                ].signal_frequency_amplitude_axis
                                            )
                                            if format_type == "fft"
                                            else object_data[2].psd,
                                        ],
                                        "x": [
                                            object_data[2].frequency_axis
                                            if format_type == "fft"
                                            else object_data[2].frequency
                                        ],
                                    }
                                ],
                            ),
                        ]
                    ),
                )
            ],
        )
    else:
        pass
    return fig


def change_utc_timezone(timestamp):
    utc = pytz.timezone("UTC")
    timezone = pytz.timezone("America/Mexico_City")
    date_time = pd.to_datetime(timestamp)
    localized_timestamp = utc.localize(date_time)
    new_timezone = localized_timestamp.astimezone(timezone)
    return new_timezone


def separate_df_per_dimension(df):
    x = df.loc[df["dimension"] == 0]
    y = df.loc[df["dimension"] == 1]
    z = df.loc[df["dimension"] == 2]

    sorted_dfs = []
    for df in [x, y, z]:
        df_drop = df.drop(columns=["dimension"])
        df_drop["observation_timestamp"] = (
            pd.to_datetime(df_drop["observation_timestamp"])
            .dt.tz_localize("UTC")
            .dt.tz_convert("America/Mexico_City")
        )
        sorted_df = df_drop.sort_values(by="observation_timestamp").reset_index(
            drop=True
        )
        sorted_dfs.append(sorted_df)
    return sorted_dfs[0], sorted_dfs[1], sorted_dfs[2]


def metrics_visualization(df):
    df = df.drop(
        columns=[
            "edge_logical_id",
            "package_timestamp",
            "perception_name",
            "signal_timestamp",
            "configuration_id",
            "partition_timestamp",
            "tenant_id",
            "dataset_versions",
        ],
        errors="ignore",
    )

    xdf, ydf, zdf = separate_df_per_dimension(df)
    column_names = list(xdf.columns.values)
    column_names.remove("observation_timestamp")

    fig = plotly.subplots.make_subplots(
        rows=len(column_names), cols=1, vertical_spacing=0.03
    )

    for index, colname in enumerate(column_names):
        fig.add_trace(
            go.Scatter(
                name=f"X-{colname}",
                x=xdf["observation_timestamp"][::-1],
                y=xdf[colname][::-1],
                visible=True,
            ),
            row=index + 1,
            col=1,
        )
        fig.add_trace(
            go.Scatter(
                name=f"Y-{colname}",
                x=ydf["observation_timestamp"][::-1],
                y=ydf[colname][::-1],
                visible=False,
            ),
            row=index + 1,
            col=1,
        )
        fig.add_trace(
            go.Scatter(
                name=f"Z-{colname}",
                x=zdf["observation_timestamp"][::-1],
                y=zdf[colname][::-1],
                visible=False,
            ),
            row=index + 1,
            col=1,
        )

    fig.update_layout(
        margin=dict(l=0, r=0, t=0, b=0),
        width=500,
        height=900,
        updatemenus=[
            dict(
                type="buttons",
                showactive=True,
                buttons=list(
                    [
                        dict(
                            label="Líneas",
                            method="restyle",
                            args=["type", "markers"],
                        ),
                        dict(
                            label="Barras",
                            method="restyle",
                            args=["type", "bar"],
                        ),
                    ]
                ),
            ),
            dict(
                type="buttons",
                showactive=True,
                yanchor="top",
                x=0.1,
                y=1.09,
                direction="right",
                pad={"r": 10, "t": 10},
                xanchor="left",
                buttons=list(
                    [
                        dict(
                            label="x",
                            method="restyle",
                            args=[{"visible": [True, False, False] * len(xdf.columns)}],
                        ),
                        dict(
                            label="y",
                            method="restyle",
                            args=[{"visible": [False, True, False] * len(xdf.columns)}],
                        ),
                        dict(
                            label="z",
                            method="restyle",
                            args=[{"visible": [False, False, True] * len(xdf.columns)}],
                        ),
                    ]
                ),
            ),
        ],
    )
    fig.update_layout(
        annotations=[
            dict(
                text="Ejes",
                x=0,
                xref="paper",
                y=1.07,
                yref="paper",
                align="left",
                showarrow=False,
            )
        ]
    )

    st.plotly_chart(fig, use_container_width=True)


def get_element_cloud_to_condition_perspective(
    tenant_id,
    vibrations_time_dataset,
    current_edge,
    current_perception,
    client,
):
    timezone = pytz.timezone("America/Mexico_City")

    last_hour = datetime.datetime.now() - datetime.timedelta(hours=1)
    localized_last_hour = timezone.localize(last_hour)

    now = timezone.localize(datetime.datetime.now())

    element = bigquery_controllers.get_last_signal_from_bigquery(
        tenant_id,
        f"{vibrations_time_dataset}.time",
        current_edge,
        current_perception,
        localized_last_hour.astimezone(pytz.utc),
        now.astimezone(pytz.utc),
        client,
    )
    return element


def plotting_cloud_data_for_condition_perspective(
    format_type,
    element,
    project_data,
    current_perception,
):

    if not element is None:
        for i in project_data["devices"]:
            if i["kind"] == "Icomox":
                sampling_frequency = i["perceptions"][current_perception][
                    "sampling_frequency"
                ]
        if len(element) == 0:
            st.info("No se registran datos")
            return None
        else:
            data = []
            for dimension in range(0, len(element["data"])):
                if format_type == "fft":
                    data_object = FFTPlotter(
                        element["data"][dimension]["values"], None, sampling_frequency
                    )

                else:
                    f, psd = metrics.get_psd(
                        element["data"][dimension]["values"], sampling_frequency
                    )
                    data_object = PSDPlotter(f, psd)
                data.append(data_object)
            plot = plot_per_condition(
                format_type=format_type,
                object_data=data,
                current_perception=current_perception,
            )

            plot.update_layout(
                margin=dict(l=0, r=0, t=0, b=0),
                height=200,
            )
        return plot
    else:
        pass


def plotting_cloud_data_for_time_perspective(
    project_data,
    current_perception,
    data,
    format_type,
    user
):
    for i in project_data.device_kinds_table.values():
        if i["kind"] == "Icomox":
            sampling_frequency = i["perceptions"][current_perception][
                "sampling_frequency"
            ]

    start = time.perf_counter()
    logging.info(f"Agregando trazos a la gráfica {format_type}")

    x_range = []
    labels = []
    plts = []
    for element in data:
        sub = []
        for dimension in range(0, len(element["data"])):
            if format_type == "fft":
                plt = FFTPlotter(
                    element["data"][dimension]["values"], None, sampling_frequency
                )
            else:
                f, psd = metrics.get_psd(
                    element["data"][dimension]["values"], sampling_frequency
                )
                plt = PSDPlotter(f, psd)
            sub.append(plt)
        plts.append(sub)
        timestamp = change_utc_timezone(element["observation_timestamp"])
        x_range.append(timestamp)
        try:
            labels_data = AnnotationLabel.parse_raw(element["label"])
            labels.append(labels_data.label_value)
        except:
            labels.append(None)

    fig = add_trace(format_type, x_range, plts, current_perception, labels, user)

    camera = dict(eye=dict(x=2, y=2, z=0.1))
    fig.update_layout(
        annotations=[dict(x=0, y=0.75, text=" ", showarrow=False)],
        autosize=False,
        margin=dict(l=0, r=0, t=0, b=0),
        width=650,
        height=650,
        scene=dict(aspectmode="manual", aspectratio=dict(x=2.5, y=1.8, z=1)),
        scene_camera=camera,
    )

    logging.info(f"Se ha agregado todos los trazos a la gráfica {format_type}")
    end = time.perf_counter()

    logging.info(f"Graficar {format_type} tardó {(end - start)}")
    return fig


def plotting_time_data(
    project_data,
    perception,
    vibration_data,
    fft_metrics_df,
    psd_metrics_df,
    user,
    vibration_charts
):
    if not vibration_data is None:
        fft = plotting_cloud_data_for_time_perspective(
            project_data, perception, vibration_data, "fft", user
        )
        vibration_charts.fft_chart = fft
        vibration_charts.fft_metrics_df = fft_metrics_df
        psd = plotting_cloud_data_for_time_perspective(
            project_data, perception, vibration_data, "psd", user
        )
        vibration_charts.psd_chart = psd
        vibration_charts.psd_metrics_df = psd_metrics_df

    else:
        st.info("No se encuentran registros para la fecha y hora seleccionada.")


def get_condition_metrics_dataframe(
    metrics_table, tenant_id, current_edge, current_perception, client
):
    timezone = pytz.timezone("America/Mexico_City")

    last_hour = datetime.datetime.now() - datetime.timedelta(hours=1)
    localized_last_hour = timezone.localize(last_hour)

    now = timezone.localize(datetime.datetime.now())

    df = bigquery_controllers.get_all_records_within_interval(
        tenant_id,
        metrics_table,
        current_edge,
        current_perception,
        localized_last_hour.astimezone(pytz.utc),
        now.astimezone(pytz.utc),
        "observation_timestamp",
        client,
    )

    return df.tail(1)


def plotting_condition_data(
    vibrations_time_dataset,
    render_card,
    project_data,
    current_perception,
    tenant_id,
    current_edge,
    client,
):
    element = get_element_cloud_to_condition_perspective(
        tenant_id,
        vibrations_time_dataset,
        current_edge,
        current_perception,
        client,
    )
    fft = plotting_cloud_data_for_condition_perspective(
        "fft",
        element,
        project_data,
        current_perception,
    )

    psd = plotting_cloud_data_for_condition_perspective(
        "psd",
        element,
        project_data,
        current_perception,
    )

    fft_metrics = get_condition_metrics_dataframe(
        f"{vibrations_time_dataset}.fft_metrics",
        tenant_id,
        current_edge,
        current_perception,
        client,
    )
    psd_metrics = get_condition_metrics_dataframe(
        f"{vibrations_time_dataset}.psd_metrics",
        tenant_id,
        current_edge,
        current_perception,
        client,
    )
    if fft_metrics.empty and psd_metrics.empty:
        pass
    else:
        strtime = change_utc_timezone(
            fft_metrics["observation_timestamp"].values[0]
        ).strftime("%d/%m/%Y, %H:%M:%S")

        fft_summary_card = Card(
            1,
            1,
            1,
            "Métricas",
            {
                "FFT": strtime,
                "Pico": fft_metrics["peak"].tolist(),
                "RMS": fft_metrics["rms"].tolist(),
                "Desviación estándar": fft_metrics["standard_deviation"].tolist(),
                "Factor de cresta": fft_metrics["crest_factor"].tolist(),
            },
            ["FFT", "Pico", "RMS", "Desviación estándar", "Factor de cresta"],
            vertical_style=False,
        )
        psd_summary_card = Card(
            1,
            1,
            3,
            "Métricas",
            {
                "PSD": pd.to_datetime(fft_metrics["observation_timestamp"]).astype(str),
                "Pico de frecuencia": psd_metrics["peak_frequency"].tolist(),
                "Pico de amplitud": psd_metrics["peak_amplitude"].tolist(),
                "RMS": psd_metrics["rms"].tolist(),
                "Desviacion estandar": psd_metrics["standard_deviation"].tolist(),
                "Factor de cresta": psd_metrics["crest_factor"].tolist(),
            },
            [
                "PSD",
                "Pico de frecuencia",
                "Pico de amplitud",
                "RMS",
                "Desviacion estandar",
                "Factor de cresta",
            ],
            vertical_style=False,
        )
    c = st.empty()
    with c.container():
        if element is not None:
            st.success("Condición del motor: Óptima")
            st.markdown("Datos registrados al momento de último cambio de condición")
            st.markdown("##### FFT")
            st.plotly_chart(fft, use_container_width=True)

            if not fft_metrics.empty:
                render_card([fft_summary_card])

            st.markdown("#### PSD")
            st.plotly_chart(psd, use_container_width=True)
            if not fft_metrics.empty:
                render_card([psd_summary_card])
        else:
            st.info("Sin data para mostrar.")



import struct
import common.common_symbols

NO_REPORT_TIMEOUT_SEC               = 0
UART_IN_MSG_PREFIX = bytes(b"KOBI")
SMIP_PACKET_HEADER_SIZE             = 1

M24C64_PAGE_BYTE_SIZE               = 32
M24C64_BYTES_SIZE                   = 8192

# eAXIS
cAXIS_X                             = 0
cAXIS_Y                             = 1
cAXIS_Z                             = 2

# eBACKBONE_PROTOCOL
cBACKBONE_PROTOCOL_TCP              = 0
cBACKBONE_PROTOCOL_MQTT             = 1

cBACKBONE_PROTOCOL_COUNT            = 1 # No MQTT right now

# eCOMOX_MSG_CODE
cCOMOX_MSG_CODE_Hello               = 0
cCOMOX_MSG_CODE_Reset               = 1
cCOMOX_MSG_CODE_GetConfiguration    = 2
cCOMOX_MSG_CODE_SetConfiguration    = 3
cCOMOX_MSG_CODE_ReadEEPROM          = 4
cCOMOX_MSG_CODE_WriteEEPROM         = 5
cCOMOX_MSG_CODE_VerifyEEPROM        = 6
cCOMOX_MSG_CODE_Debug               = 7

cCOMOX_MSG_CODE_COUNT               = 8
cCOMOX_MSG_CODE_Report              = 0xFF


# eCOMOX_RESULT
cCOMOX_RESULT_OK                                = 0
cCOMOX_RESULT_UNKNOWN_ERROR                     = 1
cCOMOX_RESULT_UNSUPPORTED_FEATURE               = 2
cCOMOX_RESULT_SD_CARD                           = 3
cCOMOX_RESULT_INVALID_EEPROM_COUNT              = 4
cCOMOX_RESULT_INVALID_EEPROM_ADDRESS            = 5
cCOMOX_RESULT_INVALID_EEPROM_ADDRESS_AND_COUNT  = 6
cCOMOX_RESULT_EEPROM_WRITE_BOUNDARY_ERROR       = 7
cCOMOX_RESULT_EEPROM_VERIFY_FAILED              = 8
cCOMOX_RESULT_EEPROM_ACCESS_IS_NOT_ALLOWED      = 9

cCOMOX_RESULT_COUNT                 = 9

# eCOMOX_BOARD_TYPR
cCOMOX_BOARD_TYPE_SMIP              = 0
cCOMOX_BOARD_TYPE_NB_IOT            = 1
cCOMOX_BOARD_TYPE_POE               = 2

cCOMOX_BOARD_TYPE_COUNT             = 3

# eCOMOX_PAYLOAD for sIN_MSG_Report.payloadType:
cCOMOX_SENSOR_ADXL362               = 0
cCOMOX_SENSOR_ADXL356               = 1
cCOMOX_SENSOR_BMM150                = 2
cCOMOX_SENSOR_ADT7410               = 3
cCOMOX_SENSOR_IM69D130              = 4
cCOMOX_SENSOR_ADXL1002              = 5

cCOMOX_SENSOR_COUNT                 = 6

# cCOMOX_KeepAlive                    = 0xFF

# uCOMOX_CONFIGURATION_ReportedPayloads
cCOMOX_SENSOR_BITMASK_ADXL362       = 1 << 0
cCOMOX_SENSOR_BITMASK_ADXL356       = 1 << 1
cCOMOX_SENSOR_BITMASK_BMM150        = 1 << 2
cCOMOX_SENSOR_BITMASK_ADT7410       = 1 << 3
cCOMOX_SENSOR_BITMASK_IM69D130      = 1 << 4
cCOMOX_SENSOR_BITMASK_ADXL1002      = 1 << 5

# uCOMOX_CONFIGURATION_CommChannel
cCOMOX_CONFIGURATION_COMM_CHANNEL_USB              = 0
cCOMOX_CONFIGURATION_COMM_CHANNEL_AUX              = 1

# uCOMOX_CONFIG_Common
cCOMOX_CONFIGURATION_Activate_CommChannelIsAux      = 1 << 0
cCOMOX_CONFIGURATION_Activate_Vibrator              = 1 << 1
cCOMOX_CONFIGURATION_Activate_Transmit              = 1 << 6
cCOMOX_CONFIGURATION_Activate_SaveToFile            = 1 << 7

# eCOMOX_RESET_TYPE
cCOMOX_RESET_TYPE_Software          = 0
cCOMOX_RESET_TYPE_Hardware          = 1
cCOMOX_RESET_TYPE_FirmwareUpdate    = 2

# eBG96_ACTIVITY_BITMASK
cBG96_ACTIVITY_BITMASK_UART         = 0x01
cBG96_ACTIVITY_BITMASK_SIM          = 0x02
cBG96_ACTIVITY_BITMASK_Ping         = 0x04

cMODULE_RawData                     = 0
cMODULE_AnomalyDetection            = 1
cMODULE_Maintenance                 = 2
cMODULE_Debug                       = 3
# cMODULE_COUNT                       = 4

cMODULE_BITMASK_RawData             = 0x01
cMODULE_BITMASK_AnomalyDetection    = 0x02
cMODULE_BITMASK_Maintenance         = 0x04

cCOMOX_CONFIG_BITMASK_AbsoluteTime          = 0x01
cCOMOX_CONFIG_BITMASK_Common                = 0x02
cCOMOX_CONFIG_BITMASK_PeriodicActivation    = 0x04

cANOMALY_DETECTION_COMMAND_Train             = 0
cANOMALY_DETECTION_COMMAND_Reset             = 1

cCOMOX_DEBUG_NBIOT_CMD_Send_AT_Command      = 0
cCOMOX_DEBUG_NBIOT_CMD_Reset_BG96           = 1
cCOMOX_DEBUG_NBIOT_CMD_TestConnectivity     = 2

DEBUG_REPOR_PAYLOAD_SIZE                    = 2048

# sCOMOX_OUT_MSG_xxxx
def OUT_MSG_Hello():
    return struct.pack("<B", cCOMOX_MSG_CODE_Hello)


def OUT_MSG_Reset(ResetType):
    return struct.pack("<BB", cCOMOX_MSG_CODE_Reset, ResetType)


def OUT_MSG_GetConfiguration():
    return struct.pack("<B", cCOMOX_MSG_CODE_GetConfiguration)


def OUT_MSG_SetConfiguration(ConfigBitmask=0, ConfigModulesBitmask=0, Common=0, LocalTimestamp=0, ActiveModules=0, RawData_Sensors=0, Repetition=0, IntervalInMinutes=0, AnomalyDetection_Command=cANOMALY_DETECTION_COMMAND_Train, AnomalyDetection_Sensors=0, AnomalyDetection_StateToTrain=0, Maintenance_Sensors=0, Maintenance_Flags=0, Maintenance_Alpha=0):
    return struct.pack("<BBBB" "q" "HB" "B" "B" "BBB" "BBH",
                       cCOMOX_MSG_CODE_SetConfiguration, ConfigBitmask, ConfigModulesBitmask, Common,
                       LocalTimestamp,
                       IntervalInMinutes, Repetition,
                       ActiveModules,
                       RawData_Sensors,
                       AnomalyDetection_Command, AnomalyDetection_Sensors, AnomalyDetection_StateToTrain,
                       Maintenance_Sensors, Maintenance_Flags, Maintenance_Alpha
                       )

# def OUT_MSG_SetConfiguration(RawDataSensors, CommChannel, Activate, Repetition, IntervalInMinutes, LocalTimestamp):
#     return struct.pack("<BBBBBHq", cCOMOX_MSG_CODE_SetConfiguration, RawDataSensors, CommChannel, Activate, Repetition, IntervalInMinutes, LocalTimestamp)


def OUT_MSG_ReadEEPROM(Count, Address):
    return struct.pack("<BBH", cCOMOX_MSG_CODE_ReadEEPROM, Count, Address)


def OUT_MSG_WriteEEPROM(Count, Address, Data):
    if len(Data) < M24C64_PAGE_BYTE_SIZE:
        Data += bytearray(M24C64_PAGE_BYTE_SIZE-len(Data))
    return struct.pack("<BBH", cCOMOX_MSG_CODE_WriteEEPROM, Count, Address) + Data


def OUT_MSG_VerifyEEPROM(Count, Address, Data):
    if len(Data) < M24C64_PAGE_BYTE_SIZE:
        Data += bytearray(M24C64_PAGE_BYTE_SIZE-len(Data))
    return struct.pack("<BBH", cCOMOX_MSG_CODE_VerifyEEPROM, Count, Address) + Data


def OUT_MSG_Debug_Send_AT_Command(AT_command_string, timeout_in_msec):
    return struct.pack("<BBH128s", cCOMOX_MSG_CODE_Debug, cCOMOX_DEBUG_NBIOT_CMD_Send_AT_Command, timeout_in_msec, bytearray(AT_command_string, "utf-8"))


def OUT_MSG_Debug_Reset_BG96():
    return struct.pack("<BB130s", cCOMOX_MSG_CODE_Debug, cCOMOX_DEBUG_NBIOT_CMD_Reset_BG96, bytearray(130))


def OUT_MSG_Debug_TestConnectivity():
    return struct.pack("<BB130s", cCOMOX_MSG_CODE_Debug, cCOMOX_DEBUG_NBIOT_CMD_TestConnectivity, bytearray(130))


# sCOMOX_IN_MSG_xxxx
def IN_MSG_Hello(msg):
    board_type, \
    board_version_major, board_version_minor = struct.unpack_from("<BBB", msg, 1)
    mcu_serial_number = msg[4:20]
    firmware_release_version_major, firmware_release_version_minor, firmware_release_version_patch, firmware_release_version_branch, \
    firmware_build_version_date_year, firmware_build_version_date_month, firmware_build_version_date_day, \
    firmware_build_version_time_hour, firmware_build_version_time_minute, firmware_build_version_time_second, \
    bit_status, \
    product_part_number, production_serial_number, name, \
    smip_swMajor, smip_swMinor, smip_swPatch, smip_swBuild = struct.unpack_from("<BBBBHBBBBBB32s32s32sBBBH", msg, 20)
    return board_type, \
           board_version_major, board_version_minor, \
           mcu_serial_number, \
           firmware_release_version_major, firmware_release_version_minor, firmware_release_version_patch, firmware_release_version_branch, \
           firmware_build_version_date_year, firmware_build_version_date_month, firmware_build_version_date_day, \
           firmware_build_version_time_hour, firmware_build_version_time_minute, firmware_build_version_time_second, \
           bit_status, \
           product_part_number, production_serial_number, name, \
           smip_swMajor, smip_swMinor, smip_swPatch, smip_swBuild


def IN_MSG_Reset(msg):
    return


def IN_MSG_GetConfiguration(msg):
    Common, \
    TransmitRepetition, TransmitIntervalInMinutes, \
    ActiveModules, \
    RawDataSensors, \
    AnomalyDetectionModule_target, AnomalyDetectionModule_selSensor, AnomalyDetectionModule, AnomalyDetectionModule_modelLearned, \
    MaintenanceModule_sensors, MaintenanceModule_flags, MaintenanceModule_alpha = \
        struct.unpack_from("<B" "BH" "B" "B" "BBBB" "BBH", msg, 1)
    return Common, TransmitRepetition, TransmitIntervalInMinutes, ActiveModules, \
            RawDataSensors, \
            AnomalyDetectionModule_target, AnomalyDetectionModule_selSensor, AnomalyDetectionModule, AnomalyDetectionModule_modelLearned, \
            MaintenanceModule_sensors, MaintenanceModule_flags, MaintenanceModule_alpha


def IN_MSG_SetConfiguration(msg):
    return msg[1]


def IN_MSG_Debug(msg):
    # [s] = struct.unpack_from("<{}s".format(DEBUG_REPOR_PAYLOAD_SIZE), msg, 1)
    # return s    #.decode("utf-8")
    Cmd, Result = struct.unpack_from("<BB", msg, 1)
    return Cmd, Result


def IN_MSG_Report(msg):
    module = msg[1] >> 6
    [timestamp] = struct.unpack_from("<q", msg, 2)
    if module == cMODULE_RawData:
        axis = (msg[1] >> 4) & 0x03
        sensor = msg[1] & 0x07
    elif module == cMODULE_Debug:
        command = (msg[1] & 0x03)
        last_packet = (msg[1] & 0x20) != 0
        return module, last_packet, command, timestamp
    else:
        axis = None
        sensor = None
    return module, sensor, axis, timestamp


def IN_MSG_REPORT_AnomalyDetection(payload):
    probState = [0] * 4
    ValAnomaly, probState[0], probState[1], probState[2], probState[3], ReportStatus, Sensors, Result = struct.unpack("<f4fBBB", payload)
    return ValAnomaly, probState, ReportStatus, Sensors, Result


# def IN_MSG_REPORT_Maintenance(payload):
    # BasicStats = [0, 0, 0, 0] * 6
    # for i in range(0, cCOMOX_SENSOR_COUNT):
    #     struct.unpack_from("<hhhh", payload, i * 8)
    # AlertsMagnitudes = struct.unpack_from("<{}h".format(cCOMOX_SENSOR_COUNT),
    #                                       2 + cCOMOX_SENSOR_COUNT * 8)
    # MotorOnTimeSec, TotalTimeSec, SyncSpeed, MotorSpeed, MinAlerts, MaxAlerts, IsMotorOn = struct.unpack_from("<LLhhBB?", msg, 2 + cCOMOX_SENSOR_COUNT * 10)
    # return BasicStats, AlertsMagnitudes,  MotorOnTimeSec, TotalTimeSec, SyncSpeed, MotorSpeed, MinAlerts, MaxAlerts, IsMotorOn


def IN_MSG_REPORT_Debug(payload):
    [s] = struct.unpack_from("<{}s".format(DEBUG_REPOR_PAYLOAD_SIZE), payload)
    return s    #.decode("utf-8")
    # debug_cmd, s = struct.unpack_from("<B{}s".format(DEBUG_REPOR_PAYLOAD_SIZE), msg, 1)
    # return debug_cmd, s    #.decode("utf-8")


# helpers for the OnGetIncomingMsgSize
COMOX_IN_MSG_SIZE =   \
(
    133,    # Hello
    1,      # Reset
    15,     # GetConfiguration
    2,      # SetConfiguration
    37,     # ReadEEPROM
    5,      # WriteEEPROM,
    5,      # VerifyEEPROM
    3       # Debug
)

COMOX_IN_MSG_REPORT_HEADER_SIZE = 10    # Report (only msg code + payload type + timestamp)

ADXL362_SAMPLES_NUM                 = 1024
ADXL356_SAMPLES_NUM_PER_PIN         = 2048
BMM150_SAMPLES_NUM                  = 512
IM69D130_SAMPLES_NUM                = 1024
ADXL1002_SAMPLES_NUM                = 2048

ADXL356_SMIP_SAMPLES_NUM_PER_PIN    = 8192

COMOX_IN_MSG_REPORT_PAYLOAD_SIZE =   \
(
    ADXL362_SAMPLES_NUM * 2 * 3,                    # ADXL362 - 16 bits array, 3 interleaved axis
    ADXL356_SAMPLES_NUM_PER_PIN * 2 * 3 * 12 // 16, # ADXL356 - 16 bits array, 3 axis, 1.5 bytes (=12 bits) each sample
    BMM150_SAMPLES_NUM * 3 * 2,                     # BMM150 - Each sample includes: Bx, By, Bz. Each is of 16 bits
    2,                                              # ADT7410 - temperature data, signed 16 bits, Q9.7 format
    IM69D130_SAMPLES_NUM * 2,                       # IM69D130 - Each sample is of 16 bits
    ADXL1002_SAMPLES_NUM * 2 * 12 // 16
)

COMOX_IN_MSG_REPORT_PAYLOAD_SIZE_adxl356_smip =   \
(
    ADXL362_SAMPLES_NUM * 2 * 3,                    # ADXL362 - 16 bits array, 3 interleaved axis
    ADXL356_SMIP_SAMPLES_NUM_PER_PIN * 2 * 12 // 16,# ADXL356 - 16 bits arrays of samples from a single axis. Each sample is of 12 bits compressed in the array.    #ADXL356_SAMPLES_NUM_PER_PIN * 2 * 12 // 16,
    BMM150_SAMPLES_NUM * 3 * 2,                     # BMM150 - Each sample includes: Bx, By, Bz. Each is of 16 bits
    2,                                              # ADT7410 - temperature data, signed 16 bits, Q9.7 format
    IM69D130_SAMPLES_NUM * 2                        # IM69D130 - Each sample is of 16 bits
)

COMOX_SENSOSRS_NAMES = \
(
    "ADXL362",
    "ADXL356",
    "BMM150",
    "ADT7410",
    "IM69D130",
    "ADXL1002"
)



from base64 import b64encode
import base64
import os
from pathlib import Path
import abc
from os import listdir
from PIL import Image
from os.path import isfile, join


class ResourceManagement(abc.ABC):
    """
    Image management to include in txp_web
    """

    def __init__(self, file_name: str):
        self.current_directory = os.path.dirname(os.path.realpath(__file__))
        self.file_name = file_name
        self.resource_name = self.get_resource_name()
        self.full_path = self._get_resource_path()
        self.resource_name = self.get_resource_name()
        self.resource64 = self._resource_to_bytes()

    @abc.abstractmethod
    def get_resource_name(self):
        pass

    def _get_resource_path(self):
        """
        get the relative path from any resource from its name.
        """
        full_path = os.path.join(self.current_directory, self.resource_name, self.file_name)
        return full_path

    @abc.abstractmethod
    def get_local_html_resource(self, width):
        """
        returns the html from a local resource as an resource-64.
        """
        pass

    @abc.abstractmethod
    def _resource_to_bytes(self):
        pass


class ImageManagement(ResourceManagement):
    """
    Image management to include in txp_web
    """

    def __init__(self, file_name: str):
        super(ImageManagement, self).__init__(
            file_name
        )

    def get_resource_name(self):
        return "images"

    def _resource_to_bytes(self):
        """
        method that converts image to bytes
        """
        img_bytes = Path(self.full_path).read_bytes()
        encoded = base64.b64encode(img_bytes).decode()
        return encoded

    def get_local_html_resource(self, width=60):
        """
        returns the html img from a local resource as an image64.
        """
        return f'<img src="data:image/jpeg;base64,{self.resource64}" width={width} align=”middle”/>'

    def get_local_html_responsive_resource(self, width=60):
        """
            returns the html img from a local resource as an image64 responsive.
            """
        return f'<img class="centrado" src="data:image/jpeg;base64,{self.resource64}" width={width} align=”middle”/>'

    def get_PIL_image_list(self, sub_directory_name):
        """
        returns the list with PIL images from a sub_directory in image resources
        """
        current_directory = os.path.dirname(os.path.realpath(__file__))
        full_path = os.path.join(current_directory, "images", sub_directory_name)
        onlyfiles = [f for f in listdir(full_path) if isfile(join(full_path, f))]
        images = []
        for i in onlyfiles:
            image = Image.open(full_path + "/" + i)
            images.append(image)
        return images


class VideoManagement(ResourceManagement):
    """
    Image management to include in txp_web
    """

    def __init__(self, file_name: str):
        super(VideoManagement, self).__init__(
            file_name
        )

    def get_resource_name(self):
        return "videos"

    def _resource_to_bytes(self):
        """
        method that converts image to bytes
        """
        with open(self.full_path, "rb") as f:
            raw_data = f.read()
            data = b64encode(raw_data).decode("utf-8")
            return data

    def get_local_html_resource(self, width):
        """
        returns the html video from a local resource as an b64 encode.
        """
        return f'<video class="centrado" alt="test" loop="true" autoplay="autoplay" controls="controls" id="vid" muted> <source src="data:video/mp4;base64,{self.resource64}" type="video/mp4" width={width}/>'


def get_resource_video_path(file):
    """
    get the relative path from any resource from its name.
    """
    current_directory = os.path.dirname(os.path.realpath(__file__))
    resources_path = os.path.join(current_directory, "videos", file)
    return resources_path

def get_resource_image_path(file):
    """
    get the relative path from any resource from its name.
    """
    current_directory = os.path.dirname(os.path.realpath(__file__))
    resources_path = os.path.join(current_directory, "images", file)
    return resources_path


"""
This module implements the left Sidebar for the layout of the application.
"""
from txp.web.core_components.app_component import AppComponent, AppState
from txp.web.core_components.app_profiles import AppProfile
from txp.web.resources.resources_management import ImageManagement
from typing import Dict, List, Callable
import streamlit as st
import logging
import os
# TODO: This might be a problem. The logging configuration is being taken from `txp` package
from txp.common.config import settings
log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


class Sidebar(AppComponent):
    """The left Sidebar component for the layout of the application.

    This component is the `Main Menu` for the user to interact with.

    This component shows Menu widgets and updates the MainComponentApp state when
    those widgets change their state. This allows to the MainComponentApp to determine
    which MainView children components to render.
    """
    def __init__(self):
        """
        Args:
            _update_profile_callback: callback to be called when the profile widget
                change its value.

            _update_selected_view: callback to be called when the selected view change
                its value.
        """
        super(Sidebar, self).__init__(
            component_key='sidebar'
        )

    @classmethod
    def _register_app_state(cls, state: AppState) -> None:
        raise NotImplemented

    def _current_profile_change(self):
        new_profile = st.session_state["current_profile_selectbox"]
        self.app_state.current_profile = AppProfile(new_profile)
        log.debug(f'Current profile changed to: {new_profile}')

    def _current_view_change(self):
        new_view = st.session_state["current_view_radio"]
        self.app_state.current_view = new_view
        log.debug(f'Current view changed to: {new_view}')

    def _render(self) -> None:
        current_directory = os.path.dirname(os.path.realpath(__file__))
        st.sidebar.image(f'{current_directory}/../resources/images/txp_logo_1.png', output_format="PNG")

        st.sidebar.selectbox(
            label="Perfil Actual",
            options=AppProfile.get_display_values(),
            key="current_profile_selectbox",
            on_change=self._current_profile_change
        )

        profile = st.session_state["current_profile_selectbox"]
        views_profile = self.app_state.views_for_profile[profile]



        last_radio_value = st.sidebar.radio(
            label="Opciones Disponibles",
            options=views_profile,
            key="current_view_radio"
        )

        self._current_view_change()  # Updates last radio value to cover all cases of interaction



import ray


@ray.remote
class ActorsStack:
    """ This class is used to keep track of running actors in the service.
    In order to shut down prediction service correctly we need to shut it down when no processes
    are running.
    This class provides 2 decorators.

    stateful_method: use this decorator whenever the method of an actor is called and you want it
    method to be stateful.

    stateful_function: use this decorator when you want to make a ray remote function stateful
    """
    def __init__(self):
        self.running_actors = 0

    def push(self):
        self.running_actors += 1

    def get_running_actors_count(self):
        return self.running_actors

    def pop(self):
        if self.running_actors:
            self.running_actors -= 1
        else:
            raise Exception("Trying to remove an actor from stack but stack is empty")


def stateful_method(func):
    def inner(self, *args, **kwargs):
        if 'actors_stack' in kwargs:
            stack = kwargs['actors_stack']
            stack.push.remote()
            kwargs.pop('actors_stack')
            try:
                ret = func(self, *args, **kwargs)
                stack.pop.remote()
            except Exception as e:
                stack.pop.remote()
                raise e
        else:
            ret = func(self, *args, **kwargs)
        return ret
    return inner


def stateful_function(func):
    def inner(*args, **kwargs):
        if 'actors_stack' in kwargs:
            stack = kwargs['actors_stack']
            stack.push.remote()
            try:
                ret = func(*args, **kwargs)
                stack.pop.remote()
            except Exception as e:
                stack.pop.remote()
                raise e
        else:
            ret = func(*args, **kwargs)
        return ret
    return inner



from airflow import DAG
from airflow.operators.dummy_operator import DummyOperator
from airflow.operators.python_operator import PythonOperator
from airflow.contrib.operators.dataflow_operator import DataFlowPythonOperator
from airflow import configuration

import google.cloud.firestore as firestore
import datetime
import os

#######################################################################################
# PARAMETROS
#######################################################################################
"""nameDAG: DAG name to be deployed """
nameDAG = 'DAG-signals-publisher'
"""trigger_time: UTC time according to https://crontab.guru/ format. e.g. 0 11 * * * = 11:00 UTC = 5:00 CST"""
trigger_time = "0 11 * * *"

default_args = {
    'depends_on_past': True,
    'start_date': datetime.datetime(2021, 11, 5),
    'email_on_failure': False,
    'email_on_retry': False,
    'retries': 2,
    'retry_delay': datetime.timedelta(minutes=1),
    'project_id': os.environ.get("GCP_PROJECT_ID", "tranxpert-mvp"),
    'firestore_pending_signal_chunks_collection': os.environ.get("PENDING_SIGNALS_CHUNKS_COLLECTION",
                                                                 "pending_signal_chunks_test"),
    'firestore_signals_queue_collection': os.environ.get("SIGNALS_QUEUE_COLLECTION",
                                                         "signals_queue_test"),
    'model_signals_topic_name': os.environ.get("MODEL_SIGNALS_TOPIC_NAME", "txp-model-serving-signals-test"),
    'dataset': os.environ.get("DATASET", "telemetry_test"),
    'data_lake': os.environ.get("DATA_LAKE", "tranxpert-mvp-telemetry-data-lake-test"),
    'composer_resources_bucket': os.environ.get("RESOURCES_BUCKET", "gs://composer-resources-test"),
    'location': os.environ.get('LOCATION')
}


LOCAL_SETUP_FILE = os.path.join(configuration.get('core', 'dags_folder'), 'setup.py')


#######################################################################################


def get_prefix_id(path):
    split_path = path.split("/")[1].split("|")
    return f"{split_path[0]}|{split_path[1]}|{split_path[2]}|{split_path[3]}"


def check_completeness(paths):
    for path in paths:
        split_path = path.split("|")
        if split_path[5] == '0':
            if (int(split_path[4]) + 1) == len(paths):
                return True
    return False


def get_complete_signals(entries):
    by_id = {}
    for entry in entries:
        if entry.to_dict():
            path = entry.get("path")
            prefix_id = get_prefix_id(path)
            if prefix_id in by_id:
                by_id[prefix_id].append(path)
            else:
                by_id[prefix_id] = [path]
    for prefix_id in by_id:
        if check_completeness(by_id[prefix_id]):
            yield by_id[prefix_id]


######################################################

def python_get_signals_from_datalake_and_push_to_queue(ds, **kwargs):
    db = firestore.Client()
    entries = db.collection(default_args['firestore_pending_signal_chunks_collection']).stream()

    paths = list(get_complete_signals(entries))
    firestore_db = firestore.Client()

    for item in paths:
        push_signal_to_signals_queue(get_prefix_id(item[0]), item, firestore_db)
        delete_documents(firestore_db, item)


def delete_documents(db, paths):
    for path in paths:
        doc = path.split("/")[1]
        db.collection(default_args['firestore_pending_signal_chunks_collection']).document(doc).delete()


def push_signal_to_signals_queue(signal_id, paths, db):
    collection = default_args['firestore_signals_queue_collection']
    db.collection(collection).document(signal_id).set({
        "paths": paths
    })


with DAG(nameDAG,
         default_args=default_args,
         catchup=False,
         max_active_runs=3,
         schedule_interval=trigger_time) as dag:
    #############################################################

    t_begin = DummyOperator(task_id="begin")

    task_get_signals_from_datalake_and_push_to_queue = PythonOperator(
        task_id='task_get_signals_from_datalake_and_push_to_queue',
        provide_context=True,
        python_callable=python_get_signals_from_datalake_and_push_to_queue
        )

    task_launch_batch_dataflow = DataFlowPythonOperator(
        task_id='task_launch_batch_dataflow',
        py_file=f'{default_args["composer_resources_bucket"]}/pipelines/batch_pipeline/batch_pipeline.py',
        gcp_conn_id='google_cloud_default',
        options={
            "job_name": 'telemetry-batch-from-composer',
            "time_table": f'tranxpert-mvp:{default_args["dataset"]}.time',
            "fft_table": f'tranxpert-mvp:{default_args["dataset"]}.fft',
            "psd_table": f'tranxpert-mvp:{default_args["dataset"]}.psd',
            "time_metrics_table": f'tranxpert-mvp:{default_args["dataset"]}.time_metrics',
            "fft_metrics_table": f'tranxpert-mvp:{default_args["dataset"]}.fft_metrics',
            "psd_metrics_table": f'tranxpert-mvp:{default_args["dataset"]}.psd_metrics',
            "model_signals_topic_name": default_args['model_signals_topic_name'],
            "setup_file": LOCAL_SETUP_FILE,
            "signals_queue_collection": default_args["firestore_signals_queue_collection"],
            "data_lake": default_args["data_lake"]
        },
        dataflow_default_options={
            "project": default_args['project_id'],
            "region": default_args["location"],
            "temp_location": 'gs://telemetry-pipeline-bucket',
        },
        dag=dag
    )

    t_end = DummyOperator(task_id="end")

    #############################################################
    t_begin >> task_get_signals_from_datalake_and_push_to_queue >> task_launch_batch_dataflow >> t_end



import setuptools

setuptools.setup(
    name='batch_signals_to_bigquery',
    version='0.1',
    description='Dependencies',
    install_requires=[
        "google-cloud-bigquery==2.30.1",
        "google==3.0.0",
        "google-cloud-firestore==2.3.4",
        "google-cloud-storage==1.43.0",
        "google-cloud-pubsub==2.11.0",
        "txp"
    ],
    packages=setuptools.find_packages()
)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gateway_package.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# This patch in the import was manually written.
# TODO: research the solution to relative imports and fix this hack.
import txp.common.protos.gateway_config_pb2 as gateway__config__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='gateway_package.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x15gateway_package.proto\x1a\x14gateway_config.proto\".\n\x1b\x44imensionSignalSamplesProto\x12\x0f\n\x07samples\x18\x01 \x03(\x02\"\xd8\x01\n\x0bSignalProto\x12-\n\x07samples\x18\x01 \x03(\x0b\x32\x1c.DimensionSignalSamplesProto\x12\x17\n\x0fperception_name\x18\x02 \x01(\t\x12\x1d\n\x15perception_dimensions\x18\x03 \x03(\x05\x12\x1d\n\x10signal_frequency\x18\x04 \x01(\x02H\x00\x88\x01\x01\x12\x1b\n\x13sampling_resolution\x18\x05 \x01(\x05\x12\x11\n\ttimestamp\x18\x06 \x01(\x03\x42\x13\n\x11_signal_frequency\"\xbf\x01\n\x1bGatewayPackageMetadataProto\x12-\n\x0f\x65\x64ge_descriptor\x18\x01 \x01(\x0b\x32\x14.EdgeDescriptorProto\x12-\n\x0fsampling_window\x18\x02 \x01(\x0b\x32\x14.SamplingWindowProto\x12\x12\n\npackage_id\x18\x03 \x01(\t\x12\x1b\n\x13previous_part_index\x18\x04 \x01(\x05\x12\x11\n\ttenant_id\x18\x0b \x01(\t\"~\n\x13GatewayPackageProto\x12.\n\x08metadata\x18\x01 \x01(\x0b\x32\x1c.GatewayPackageMetadataProto\x12\x1d\n\x07signals\x18\x02 \x03(\x0b\x32\x0c.SignalProto\x12\x18\n\x10\x63onfiguration_id\x18\x03 \x01(\tb\x06proto3'
  ,
  dependencies=[gateway__config__pb2.DESCRIPTOR,])




_DIMENSIONSIGNALSAMPLESPROTO = _descriptor.Descriptor(
  name='DimensionSignalSamplesProto',
  full_name='DimensionSignalSamplesProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='samples', full_name='DimensionSignalSamplesProto.samples', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=47,
  serialized_end=93,
)


_SIGNALPROTO = _descriptor.Descriptor(
  name='SignalProto',
  full_name='SignalProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='samples', full_name='SignalProto.samples', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='perception_name', full_name='SignalProto.perception_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='perception_dimensions', full_name='SignalProto.perception_dimensions', index=2,
      number=3, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='signal_frequency', full_name='SignalProto.signal_frequency', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sampling_resolution', full_name='SignalProto.sampling_resolution', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='SignalProto.timestamp', index=5,
      number=6, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_signal_frequency', full_name='SignalProto._signal_frequency',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=96,
  serialized_end=312,
)


_GATEWAYPACKAGEMETADATAPROTO = _descriptor.Descriptor(
  name='GatewayPackageMetadataProto',
  full_name='GatewayPackageMetadataProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='edge_descriptor', full_name='GatewayPackageMetadataProto.edge_descriptor', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sampling_window', full_name='GatewayPackageMetadataProto.sampling_window', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='package_id', full_name='GatewayPackageMetadataProto.package_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='previous_part_index', full_name='GatewayPackageMetadataProto.previous_part_index', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tenant_id', full_name='GatewayPackageMetadataProto.tenant_id', index=4,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=315,
  serialized_end=506,
)


_GATEWAYPACKAGEPROTO = _descriptor.Descriptor(
  name='GatewayPackageProto',
  full_name='GatewayPackageProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metadata', full_name='GatewayPackageProto.metadata', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='signals', full_name='GatewayPackageProto.signals', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='configuration_id', full_name='GatewayPackageProto.configuration_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=508,
  serialized_end=634,
)

_SIGNALPROTO.fields_by_name['samples'].message_type = _DIMENSIONSIGNALSAMPLESPROTO
_SIGNALPROTO.oneofs_by_name['_signal_frequency'].fields.append(
  _SIGNALPROTO.fields_by_name['signal_frequency'])
_SIGNALPROTO.fields_by_name['signal_frequency'].containing_oneof = _SIGNALPROTO.oneofs_by_name['_signal_frequency']
_GATEWAYPACKAGEMETADATAPROTO.fields_by_name['edge_descriptor'].message_type = gateway__config__pb2._EDGEDESCRIPTORPROTO
_GATEWAYPACKAGEMETADATAPROTO.fields_by_name['sampling_window'].message_type = gateway__config__pb2._SAMPLINGWINDOWPROTO
_GATEWAYPACKAGEPROTO.fields_by_name['metadata'].message_type = _GATEWAYPACKAGEMETADATAPROTO
_GATEWAYPACKAGEPROTO.fields_by_name['signals'].message_type = _SIGNALPROTO
DESCRIPTOR.message_types_by_name['DimensionSignalSamplesProto'] = _DIMENSIONSIGNALSAMPLESPROTO
DESCRIPTOR.message_types_by_name['SignalProto'] = _SIGNALPROTO
DESCRIPTOR.message_types_by_name['GatewayPackageMetadataProto'] = _GATEWAYPACKAGEMETADATAPROTO
DESCRIPTOR.message_types_by_name['GatewayPackageProto'] = _GATEWAYPACKAGEPROTO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DimensionSignalSamplesProto = _reflection.GeneratedProtocolMessageType('DimensionSignalSamplesProto', (_message.Message,), {
  'DESCRIPTOR' : _DIMENSIONSIGNALSAMPLESPROTO,
  '__module__' : 'gateway_package_pb2'
  # @@protoc_insertion_point(class_scope:DimensionSignalSamplesProto)
  })
_sym_db.RegisterMessage(DimensionSignalSamplesProto)

SignalProto = _reflection.GeneratedProtocolMessageType('SignalProto', (_message.Message,), {
  'DESCRIPTOR' : _SIGNALPROTO,
  '__module__' : 'gateway_package_pb2'
  # @@protoc_insertion_point(class_scope:SignalProto)
  })
_sym_db.RegisterMessage(SignalProto)

GatewayPackageMetadataProto = _reflection.GeneratedProtocolMessageType('GatewayPackageMetadataProto', (_message.Message,), {
  'DESCRIPTOR' : _GATEWAYPACKAGEMETADATAPROTO,
  '__module__' : 'gateway_package_pb2'
  # @@protoc_insertion_point(class_scope:GatewayPackageMetadataProto)
  })
_sym_db.RegisterMessage(GatewayPackageMetadataProto)

GatewayPackageProto = _reflection.GeneratedProtocolMessageType('GatewayPackageProto', (_message.Message,), {
  'DESCRIPTOR' : _GATEWAYPACKAGEPROTO,
  '__module__' : 'gateway_package_pb2'
  # @@protoc_insertion_point(class_scope:GatewayPackageProto)
  })
_sym_db.RegisterMessage(GatewayPackageProto)


# @@protoc_insertion_point(module_scope)



from enum import Enum
from pydantic import BaseModel
from typing import Union


class ModelFeedback(BaseModel):
    pass


class ErrorModelFeedback(ModelFeedback):
    error_message: str


class SuccessModelFeedback(ModelFeedback):
    metrics: dict


class DefaultModelFeedback(ModelFeedback):
    pass


class ModelMetadata(BaseModel):
    feedback: Union[ErrorModelFeedback, SuccessModelFeedback, DefaultModelFeedback]
    tenant_id: str
    machine_id: str
    task_id: str


class ModelStateValue(Enum):
    ACKNOWLEDGE = "acknowledge"
    ACTIVE = "active"
    OLD = "old"
    TRAINING = "training"
    ERROR = "error"


class ModelState(BaseModel):
    value: ModelStateValue
    creation_date: str
    publishment_date: str
    deprecation_date: str


class ModelRegistry(BaseModel):
    metadata: ModelMetadata
    file_path: str
    state: ModelState



"""
This module contains the concrete implementation case of Driver for the
smartmesh dongle USB device.
Technical details about the device can be found in the following document:
https://tranxpert.atlassian.net/wiki/spaces/TD/pages/230817807/Tranxpert+MVP+Monitoring+Proposal#Several-Voyager-3-working-together.
"""

# ============================ imports =========================================
import queue
import time
import serial.tools.list_ports
from typing import Dict, Tuple, List

from IpMgrConnectorSerial.IpMgrConnectorSerial import IpMgrConnectorSerial
from IpMgrConnectorMux.IpMgrSubscribe import IpMgrSubscribe
from SmartMeshSDK import ApiException
from txp.devices.drivers.driver import DriverState
from txp.common.config import settings
import logging

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)

# ==============================================================================
# Voyager SmartMesh Manager Singleton
# ==============================================================================
class SmartMeshDongle(object):
    """The USB smartmesh dongle used as manager to handle smartmesh devices.

    This class is implemented following a singleton pattern.

    TODO: Custom Exception handling in the methods of the class.

    The main responsibility for this class is to communicate with the
    edge devices connected in the smartmesh network through the
    direct interaction with the smartmesh dongle device.
    """

    """
    Constant data configured values
    """
    Vcc = 3.3  # Supply Voltage
    V_perg_ref = 0.04  # V per g at reference voltage
    Vref = 1.8  # 1.8V reference voltage
    ADC_res = 16  # No of bits resolution
    ADC_offset = 2 ** (ADC_res - 1)
    V_per_LSB = Vcc / 2 ** ADC_res  # V per LSB resolution
    V_perg_Vcc = V_perg_ref * Vcc / Vref  # Volts per g at Vcc
    ADC_gain = V_per_LSB / V_perg_Vcc  # Gain of ADC
    f_adc = 6.5e6  # ADC clock frequency
    n_acq = 185  # Acq. time in clock periods

    def __init__(self, on_driver_state_change: callable):
        self.ID = 0
        self.connector = IpMgrConnectorSerial()
        self.mac = []
        """mac: The MAC address of the dongle device.
                Is this address returned in the connector call? 
                Probably some API to get it
        """
        self.operational_macs: List[Tuple] = []
        """operational_macs: A list of found operational devices MAC addresses."""
        self.drivers_packets_queues: Dict[Tuple, queue.PriorityQueue] = {}
        """edge_descriptors: A mapping of MAC address to Driver's packets queue."""
        self.num_motes: int = 0
        self.connected: bool = False
        self.net_config = None

        self.subscriber = None
        self.lights = []

        # Network Config Information
        self.netID = None
        self.max_drivers = None
        self.base_bw = None
        self.UDP_PORT_NUMBER = 61624

        # Default attributes for sampling parameters
        self.adc_num_samples = 512
        self.adc_param_len = 2

        self._port = None
        """_port: The connection port of the device."""

        #connection state change callable handler
        self.on_driver_state_change = on_driver_state_change

    def detect(self) -> bool:
        """This method looks for SmartMesh devices connected to the Gateway.

        TODO: Store the state of found ports. How to decide which port to use?

        Returns:
            True if the device was found. False otherwise.
        """
        devices = []
        ports = serial.tools.list_ports.comports()
        ports.sort(reverse=True)

        log.debug(
            f"SmartMesh manager. Ports to try connection: { [port.device for port in ports] }"
        )

        for port in ports:
            if (devices and port.vid == last_device) or not port.vid:
                continue
            try:
                log.debug(f"Trying to connect to path: {port.device} --- Device hwid: {port.hwid}")
                self.connector.connect({"port": str(port.device)})
                last_device = port.vid
                devices.append(str(port.device))
                self.connector.disconnect()
            except:
                pass
        # TODO: Selecting the first port found, based on example. Not final implementation.
        if devices:
            self._port = devices[0]
            return True

        return False

    def connect(self):
        """Connect the Manager to the SmartMesh network and the dongle device.

        If the Manager connects successfully, then it'll hold the list of operational
        edge devices by MAC addresses in self._operational_macs.

        Returns:
            None.
        """
        if self.connected:
            return
        if self._port is None:
            log.debug(
                "Trying to connect VoyagerSmartMeshManager but no connected device port was found"
            )
            raise RuntimeError(
                "Trying to connect VoyagerSmartMeshManager but no connected device port was found"
            )
        try:
            self.connector.connect({"port": self._port})
            log.info("Connected to SmartMeshNetwork")
            self.connected = True
        except:
            raise Exception("The SmartMesh manager device was not detected.")

        self.net_config = self.connector.dn_getNetworkConfig()
        self._get_motes()
        # Subscribe notification data callback
        self._subscribe_to_mote_data()

    def disconnect(self):
        """Disconnect the Manager from the SmartMesh network and dongle device.

        Returns:
             None.
        """
        if not self.connected:
            return

        self.connector.disconnect()
        self.operational_macs.clear()
        self.connected = False
        self.num_motes = 0
        log.info("SmartMesh manager was disconnected")

    def get_motes_connected(self):
        """Get the amount of motes currently connected to the manager.

        Returns:
            The amount of motes connected to the manager.
        """

        return self.num_motes

    def get_adc_number_of_samples(self):
        """Get the amount of samples taken by the ADC

        Returns:
            The amount of samples taken by the ADC
        """
        return self.adc_num_samples

    def get_adc_length(self):
        """Get the ADC length

        Returns:
            The the ADC length in bits
        """
        return self.adc_param_len

    def register_driver(
            self, mac_address: Tuple, driver_samples_queue: queue.PriorityQueue
    ):
        """Registers the Driver instance in the Manager.

        Right now, this means to provide the Manager with the priority queue
        of the driver. The manager will put received packages in the queues
        associated to the MAC address.
        """
        if mac_address in self.operational_macs:
            self.drivers_packets_queues[mac_address] = driver_samples_queue

    def unregister_device(self, mac_address: Tuple):
        """Unregister the Driver instance from the manager.

        This means to remove the driver packets queue from the manager.
        The driver will not register new packets received from the physical layer.
        """
        if mac_address in self.operational_macs:
            self.drivers_packets_queues.pop(mac_address)

    def send_to_mote(self, mac: Tuple, message: any):
        """Sends message to be received by mote device.
        # TODO: Raise exception on error or set state?
        Message can be read in mote module SmartMeshRFcog.c in RCV section.
        """
        try:
            rc = self.connector.dn_sendData(
                macAddress=mac,
                priority=0,

                srcPort=self.UDP_PORT_NUMBER,
                dstPort=self.UDP_PORT_NUMBER,
                options=0,
                data=[ord(i) for i in message],
            )
            if rc[0] == 0:
                log.info(f"Sent command to {self.pretty_mac(mac)} successfully")
            else:
                log.info(f"Unsuccessful command to {self.pretty_mac(mac)}")
        except:
            log.error(f"ERROR on send. Mote not connected. {self.pretty_mac(mac)}")

    def send_to_all_motes(self, message: any):
        """Sends message to all motes connected to the current network"""
        if self.num_motes > 0:
            log.info(f"Sending message to all motes: {message}")
            for mac in self.operational_macs:
                self.send_to_mote(mac, message)

    def _get_motes(self):
        """Asks for the connected edge devices in the network, and stores the operational
        MAC's
        """
        continueAsking = True
        currentMac = (0, 0, 0, 0, 0, 0, 0, 0)

        while continueAsking:  # While mote remains, add mote MAC to operational_macs
            try:
                res = self.connector.dn_getMoteConfig(currentMac, True)
            except ApiException.APIError:
                continueAsking = False
            else:
                if not res.isAP and res.state in [
                    4,
                ]:  # Not access point and in operational state
                    self.operational_macs.append(tuple(res.macAddress))
                currentMac = res.macAddress

        self.num_motes = len(self.operational_macs)

        for mac in self.operational_macs:
            log.info("Mote detected: {}".format(self.pretty_mac(mac)))
            self.on_driver_state_change(mac, DriverState.CONNECTED)

    def _subscribe_to_mote_data(self):
        """Subscribe the method to be called when notification data arrives from
        a mote.

        Returns:
            None
        """
        self.subscriber = IpMgrSubscribe(self.connector)
        self.subscriber.start()
        self.subscriber.subscribe(
            notifTypes=[
                IpMgrSubscribe.NOTIFDATA,  # Subscribe to notification data only
            ],
            fun=self._handle_notifdata,
            isRlbl=False,
        )

        self.subscriber.subscribe(
            notifTypes=[
                IpMgrSubscribe.NOTIFEVENT,
            ],
            fun=self._handle_notification_event,
            isRlbl=False,
        )

    def _handle_notification_event(self, notifName, notifParams):
        """Handle smartmesh manager events related to the status of the motes. Those messages can be used to manage
        the connection/disconnection events.

        Args:
            notifName: Notification ID sent by the smartmesh dongle:
            - eventMoteJoin: Mote attempting to join to the network.
            - eventMoteOperational: Mote has joined to the network.
            - eventMoteLost: Mote unreacheble or disconnected.
            notifParams: Structure holding smartmesh message metadata.
        Returns:
            None.
        """
        mac = tuple(notifParams.macAddress)
        pretty_address = self.pretty_mac(mac)
        if notifName == IpMgrSubscribe.EVENTMOTEOPERATIONAL:
            log.info("The Device with MAC {} was connected".format(pretty_address))
            self.on_driver_state_change(mac, DriverState.CONNECTED)
            if not self.operational_macs.__contains__(mac):
                self.operational_macs.append(mac)
                log.info("MAC {} has been added to list".format(pretty_address))



        if notifName == IpMgrSubscribe.EVENTMOTELOST:
            log.info("The Device with MAC {} was disconnected.".format(pretty_address))
            self.on_driver_state_change(mac, DriverState.DISCONNECTED)
            if self.operational_macs.__contains__(mac):
                self.operational_macs.remove(mac)
                log.info("MAC {} has been removed from list".format(pretty_address))


    def _handle_notifdata(self, notifName, notifParams):
        """Parses mote packet for timestamp and data. Finally, it pass
        the information to the Driver mote packets queue.

        Args:
            notifName: ID data from header of smartmesh packet, specifying type of data:
                Notification.
            notifParams: Structure holding smartmesh message data and metadata, time sent,
                and data.
        Returns:
            None.
        """
        mac_tuple = tuple(notifParams.macAddress)
        if mac_tuple not in self.drivers_packets_queues:
            return

        timestamp = notifParams.utcSecs + 1e-6 * notifParams.utcUsecs
        # Put the package in the driver priority queue, with the priority set to timestamp
        self.drivers_packets_queues[mac_tuple].put((timestamp, notifParams.data))

    @staticmethod
    def pretty_mac(mac):
        """Parses mote mac tuple into is corresponding string

        Args:
            mac: mac id tuple
        Returns:
            Mac string.
        """
        pMac = "-".join(["%02x" % b for b in mac])
        return pMac



# =========================== Imports ===============================
import base64
import sys
from txp.cloud import (
    pull_current_configuration_from_firestore,
    pull_docs_from_collection_associated_with_configuration
)
import google.cloud.firestore as firestore
from google.cloud import bigquery
from txp.cloud import settings as cloud_settings
from typing import List, Dict
from datetime import datetime
import json

# =========================== Constant Data & helper functions ===============================
edges_collection_name = cloud_settings.firestore.edges_collection
edges_states_table_name = "statesdataset01.edges_states"


def _is_gateway_state(event_data) -> bool:
    """Returns true if the event was sent by a Gateway"""
    return 'edges' in event_data


def _get_configuration_reference(db: firestore.Client, tenant_id: str) -> firestore.DocumentReference:
    configuration = pull_current_configuration_from_firestore(db, tenant_id)
    if not configuration:
        print("No Configuration document was found", file=sys.stderr)
        return None

    return configuration.reference


def _update_edges_documents(edges_new_states: Dict, documents: List[firestore.DocumentSnapshot]):
    for document in documents:
        doc_ref = document.reference
        doc_ref.update({
            'state': edges_new_states[document.get('logical_id')]
        })
        print(f'Firestore entry updated for edge: {document.get("logical_id")}')


def _clear_resources(db: firestore.Client):
    db.close()


def persist_states(client: bigquery.Client, edges_state_dict):
    table = f"""{client.project}.{edges_states_table_name}"""
    for edge_logical_id in edges_state_dict:
        select_query = f"""
        SELECT * FROM `{table}`
        WHERE edge_logical_id = "{edge_logical_id}"
        ORDER BY timestamp DESC
        LIMIT 1;
        """
        df = (client.query(select_query).result().to_dataframe())
        last_state = None
        for index, row in df.iterrows():
            last_state = row
            break
        rows_to_insert = []
        if last_state is None or edges_state_dict[edge_logical_id]["driver_state"] != last_state["state"]:
            new_state = edges_state_dict[edge_logical_id]["driver_state"]
            timestamp = edges_state_dict[edge_logical_id]["since_datetime"]
            timestamp = int(datetime.strptime(timestamp, '%Y-%m-%d %H:%M:%S').timestamp() * 1e9)
            rows_to_insert.append({"edge_logical_id": edge_logical_id, "state": new_state, "timestamp": timestamp})
        if len(rows_to_insert):
            errors = client.insert_rows_json(table, rows_to_insert)
            if not errors:
                print("New rows have been added.")
            else:
                print("Encountered errors while inserting state rows: {}".format(errors))
        else:
            print("There are not new states")


# =========================== Cloud function def ===============================
def process_state_event_to_firestore(event, context):
    """This function is the function deployed to run in the cloud
    functions runtime

    A detailed page with all the context can be found here:
        https://cloud.google.com/functions/docs/calling/pubsub
    """
    print('Starting to process event received...')
    if 'data' in event:
        state_data = base64.b64decode(event['data']).decode()
        attributes = event.get('attributes', None)
        if attributes:
            db = firestore.Client()
            bigquery_client = bigquery.Client()
            event_device_id = attributes['deviceId']
            print(f'Event received from {event_device_id}')

            if _is_gateway_state(state_data):
                edges_state_dict = json.loads(state_data)['edges']
                tenant_id = json.loads(state_data)['tenant_id']
                configuration_ref = _get_configuration_reference(db, tenant_id)
                if not configuration_ref:
                    _clear_resources(db)
                    exit(1)

                try:
                    persist_states(bigquery_client, edges_state_dict)
                except:
                    print("Error while updating state on BigQuery table", file=sys.stderr)

                print(f'State received for devices. '
                      f'This will update Firestore configuration {configuration_ref.get().get("configuration_id")}')

                # Get edges from firestore
                edges_docs: List[firestore.DocumentSnapshot] = pull_docs_from_collection_associated_with_configuration(
                    db, edges_collection_name, configuration_ref
                )
                # Filter the edges to the reported ones
                edges_docs = list(filter(
                    lambda edge_snapshot: edge_snapshot.get('logical_id') in edges_state_dict.keys(),
                    edges_docs
                ))
                # Update the edges in Firestore
                _update_edges_documents(edges_state_dict, edges_docs)

                print("The edges were updated successfully in the Database snapshot")
                _clear_resources(db)

            else:
                print('State event received from a Edge. Ignore.')

    else:
        print('Unexpected error: No data received in function.', file=sys.stderr)



from typing import Dict, List

from txp.ml.prediction_service.state_managers import state_manager as sm
from txp.common.ml.tasks import AssetTask
import logging
import datetime
import ray
from txp.common.ml.tasks import AssetState, AssetStateCondition


class BasicGaoAsset(sm.StateManager):

    def __init__(self, asset_id, tasks, events_and_states_dataset, backup_collection_name, asset_tasks: List[AssetTask],
                 notifications_topic, credentials_str, log_level=logging.INFO):
        logging.basicConfig(level=log_level)
        super().__init__(asset_id, events_and_states_dataset, backup_collection_name, notifications_topic,
                         credentials_str)
        self.tasks = tasks
        self._last_state_value: AssetStateCondition = AssetStateCondition.UNDEFINED
        self._last_critical_by_operative_flag: bool = False
        self._last_operative_checkpoint = None
        self._operative_max_allowed_secs = 600

        self._tasks_definitions: Dict[str, AssetTask] = {}
        for t in asset_tasks:
            self._tasks_definitions[t.task_id] = t

    def _get_state(self, ids: List, events: List, state_value: AssetStateCondition) -> AssetState:
        state = AssetState(events=ids, asset_id=self.asset_id, condition=state_value,
                           observation_timestamp=events[0]["observation_timestamp"], tenant_id=events[0]["tenant_id"],
                           partition_timestamp=events[0]["partition_timestamp"])
        logging.info(f"{self.__class__.__name__} computed new state for asset {state.asset_id}. "
                     f"New state value: {state.condition.value} ")
        return state

    def get_state(self, events) -> AssetState:
        # TODO: returning mock state
        logging.info(f"{self.__class__.__name__} received new events for asset {events[0]['asset_id']}.")
        rcv_events_labels: Dict[str, int] = {}

        for event in events:
            rcv_events_labels[event['task_id']] = (
                self._tasks_definitions[event['task_id']].label_def.label_to_int[event['event']]
            )

        ids: List[str] = list(map(
            lambda event_obj: event_obj['event_id'],
            events
        ))

        TP_present = (
                rcv_events_labels['TP_01_SD'] + rcv_events_labels['TP_02_MD'] + rcv_events_labels['TP_03_ID']
        )

        SD_present = (
                rcv_events_labels['SD_01_SC'] + rcv_events_labels['SD_02_SC']
        )

        EMP_present = (
                rcv_events_labels['EMP_01_SI'] + rcv_events_labels['EMP_02_MI']
        )

        first_eval = TP_present + SD_present + EMP_present

        if first_eval >= 3 and TP_present and SD_present and EMP_present:
            """First strategy, we evaluate: 

               TP:
                 TP_01_SD, TP_02_MD, TP_02_ID

               SD:
                 SD_01_SC, SD_02_MC

               EMP: 
                EMP_01_SI
                EMP_02_MI

            If at least 1 perception on each of these groups is Present, 
            then the state is optimal.
            """
            logging.info('The new state is OPTIMAL, because there is profile in TP, SD and EMP')
            new_state = self._get_state(ids, events, AssetStateCondition.OPTIMAL)

        elif TP_present:
            """At this point, we'll look for Good state.

            Good state will be truth for any Present value in the group TP. 
            TP:
                 TP_01_SD, TP_02_MD, TP_02_ID
            """
            logging.info('The new state is Good, because there are packages on TP, but not '
                         'enough in EMP and SD.')
            new_state = self._get_state(ids, events, AssetStateCondition.GOOD)

        # let's check for the last operative checkpoint time difference
        elif SD_present or EMP_present:
            """In this case, we'll look for a Operative state. 
            
            It's operative if there are profile in SD or EMP. 
            
            But we'll be careful to look if Operative tolerance has been surpassed.
            """
            logging.info('The new state is Operative, because there are packages on SD or EMP, but not '
                         'enough for a better state.')

            if self._is_valid_operative():
                logging.info('The new state is Operative, because there are packages on SD or EMP')
                self._last_operative_checkpoint = datetime.datetime.now()
                new_state = self._get_state(ids, events, AssetStateCondition.OPERATIONAL)

            else:
                logging.info('The new state is CRITICAL, because Operative was detected but the Operative tolerance '
                             'was surpassed')
                self._last_critical_by_operative_flag = True
                new_state = self._get_state(ids, events, AssetStateCondition.CRITICAL)

        else:
            logging.info('The new state is CRITICAL, because all spots are empty of profiles')
            self._last_critical_by_operative_flag = False
            new_state = self._get_state(ids, events, AssetStateCondition.CRITICAL)

        self._last_state_value = new_state.condition
        return new_state

    def _is_valid_operative(self) -> bool:
        if self._last_state_value.name != 'OPERATIONAL' and self._last_state_value.name != 'CRITICAL':
            return True

        elif self._last_state_value.name == 'CRITICAL':
            return not self._last_critical_by_operative_flag

        else:
            return (datetime.datetime.now() - self._last_operative_checkpoint).total_seconds() < \
                   self._operative_max_allowed_secs


@ray.remote
class RayBasicGaoAsset(BasicGaoAsset):
    def __init__(self, asset_id, tasks, events_and_states_dataset, backup_collection_name,
                 asset_tasks: List[AssetTask], notifications_topic, credentials_str, log_level=logging.INFO):
        super().__init__(asset_id, tasks, events_and_states_dataset, backup_collection_name, asset_tasks,
                         notifications_topic, credentials_str, log_level)



import threading
import pprint
import serial
import traceback
import os

#============================ decide whether can run ==========================

CLIB_PATHS = [
    os.path.join('libdn_ipmg'),
    os.path.join('..','..','..','sm_clib','sm_clib','libdn_ipmg'),
]
if os.name!='nt':
    for P in CLIB_PATHS:
        P += '.so'

CLIB_PATH = None
for P in CLIB_PATHS:
    if os.name=='nt':
        P = P + '.dll'
    if os.path.exists(P):
        CLIB_PATH = P
        break

if not CLIB_PATH:
    raise ImportError('Could not find C library')

#============================ logging =========================================

import logging
class NullHandler(logging.Handler):
    def emit(self, record):
        pass
log = logging.getLogger('IpMgrConnectorSerial')
log.setLevel(logging.ERROR)
log.addHandler(NullHandler())

#============================ imports =========================================

from ctypes import *

from SmartMeshSDK.utils         import FormatUtils
from SmartMeshSDK.ApiException  import ConnectionError
from SmartMeshSDK.ApiConnector  import ApiConnector
from SmartMeshSDK.ApiDefinition import ApiDefinition, \
                                       IpMgrDefinition

#============================ globals =========================================

DN_SERIAL_ST_DISCONNECTED         = 0x00
DN_SERIAL_ST_HELLO_SENT           = 0x01
DN_SERIAL_ST_CONNECTED            = 0x02

DN_ERR_NONE                       = 0

NUM_SERIAL_TRIES                  = 6

uart_txByte_cbt                   = CFUNCTYPE(c_int, c_uint8)
uart_txByte                       = None

ipmg_notif_cbt                    = CFUNCTYPE(c_void_p, c_uint8, c_uint8)
ipmg_notif                        = None

ipmg_reply_cbt                    = CFUNCTYPE(c_void_p, c_uint8)
ipmg_reply                        = None

ipmg_status_cbt                   = CFUNCTYPE(c_int, c_uint8)
ipmg_status                       = None

INT                               = IpMgrDefinition.IpMgrDefinition.INT
INTS                              = IpMgrDefinition.IpMgrDefinition.INTS
BOOL                              = IpMgrDefinition.IpMgrDefinition.BOOL
HEXDATA                           = IpMgrDefinition.IpMgrDefinition.HEXDATA

SUBID1                            = ApiDefinition.ApiDefinition.SUBID1
SUBID2                            = ApiDefinition.ApiDefinition.SUBID2

#============================ helpers =========================================

pp = pprint.PrettyPrinter(indent=4)

def buf2int(valBuf):
    valInt  = 0
    for i in range(0,len(valBuf)):
        valInt += valBuf[              i]<<(8*(len(valBuf)-1-i))
    return valInt

def _generateCParam(type,length,value):
    
    returnVal = None
    
    if   type==INT:
        returnVal       = value
    elif type==INTS:
        returnVal       = value
    elif type==BOOL:
        if   value==True:
            returnVal   = 0x01
        elif value==False:
            returnVal   = 0x00
        else:
            raise SystemError()
    elif type==HEXDATA:
        if len(value)<=4 and length!=None:
            returnVal   = buf2int(value)
        else:
            buf         = create_string_buffer(len(value))
            for (i,c) in enumerate(value):
                buf[i]  = chr(c)
            returnVal   = byref(buf)
    else:
        raise NotImplementedError('type={0}'.format(type))
    
    assert returnVal!=None
    
    return returnVal

def _getResponseType(type,length):
    assert length!=None
    if   type==INT:
        if   length==1:
            return c_uint8
        elif length==2:
            return c_uint16
        elif length==4:
            return c_uint32
        elif length==8:
            return c_uint8*8
        else:
            raise NotImplementedError
    elif type==INTS:
        if   length==1:
            return c_int8
        else:
            raise NotImplementedError
    elif type==BOOL:
        assert length==1
        return c_bool
    elif type==HEXDATA:
        return c_uint8*length

#============================ body ============================================

class IpMgrConnectorSerial(ApiConnector,threading.Thread):
    
    CONNECT_TIMEOUT               = 2
    RESPONSE_TIMEOUT              = 2
    MAX_FRAME_LENGTH              = 128
    _BAUDRATE                     = 115200
    
    def __init__(self):
        
        # local variables
        self.waitForConnected     = threading.Event()
        self.waitForResponse      = threading.Event()
        
        # initialize parent classes
        ApiConnector.__init__(self)
        threading.Thread.__init__(self)
        self.name                 = 'IpMgrConnectorSerial'
    
    def connect(self,connectParams):
        log.debug("connect connectParams={0}".format(connectParams))
        
        if 'port' not in connectParams:
            output = "'port' entry required in connection parameters"
            log.error(output)
            raise ValueError(output)
        
        #==== connect to serial port
        
        self.comPort         = connectParams['port']
        try:
            self.pyserialHandler = serial.Serial(self.comPort,baudrate=self._BAUDRATE)
            self.pyserialHandler.setRTS(False)
            self.pyserialHandler.setDTR(True)
        except serial.serialutil.SerialException as err:
            output = "could not open " + self.comPort + ", reason: " + str(err)
            log.warning(output)
            raise ConnectionError(output)
        else:
            log.info("opened port "+self.comPort)
        
        #==== import and initialize the C library
        
        global uart_txByte
        global ipmg_notif
        global ipmg_reply
        global ipmg_status
        
        # local variables
        uart_txByte               = uart_txByte_cbt(self._uart_txByte)
        ipmg_notif                = ipmg_notif_cbt(self._ipmg_notif)
        ipmg_reply                = ipmg_reply_cbt(self._ipmg_reply)
        ipmg_status               = ipmg_status_cbt(self._ipmg_status)
        
        # configure the callback function
        self.notifBuf             = create_string_buffer(self.MAX_FRAME_LENGTH)
        
        # open the library under test
        self.smipmg               = CDLL(CLIB_PATH)
        self.smipmg.dn_ipmg_init(
            ipmg_notif,           # notifCb
            self.notifBuf,        # notifBuf
            len(self.notifBuf),   # notifBufLen
            ipmg_reply,           # replyCb
            ipmg_status,          # statusCb
        )
        
        # install callbacks
        self.smipmg.dn_uart_register_txByte(uart_txByte)
        
        # initial thread
        threading.Thread.__init__(self)
        self.name = 'IpMgrConnectorSerial'
        
        # start myself
        self.start()
        
        #==== connect to manager
        
        rcApi = self.smipmg.dn_ipmg_initiateConnect()
        assert rcApi == DN_ERR_NONE
        
        # wait for C library to connect
        if self.waitForConnected.wait(self.CONNECT_TIMEOUT):
            self.waitForConnected.clear()
            # connect the parent class
            ApiConnector.connect(self)
        else:
            raise ConnectionError("timeout connecting")
    
    def run(self):
        while True:
            b = self.pyserialHandler.read(1)
            if not b:
                break
            b = ord(b)
            log.debug('RX serial byte 0x{0:02x}'.format(b))
            self.smipmg.dn_uart_rxByte(b)
        
        log.info("stopping thread")
    
    #======================== public ==========================================
    
    def send(self,commandArray,fields):
        log.debug('send commandArray={0} fields={1}'.format(commandArray,fields))
        
        assert len(commandArray)==1
        commandName = commandArray[0]
        
        apiCommand = None
        for c in IpMgrDefinition.IpMgrDefinition.commands:
            if c['name']==commandName:
                apiCommand = c
                break
        assert apiCommand
        
        #===== send command to manager
        
        # create params on the fly
        params                    = []
        varSizeParamLen           = None
        
        for [paramname,type,length,_] in apiCommand['request']:
            thisparam = {}
            thisparam['name']     = paramname
            thisparam['C']        = _generateCParam(type,length,fields[paramname])
            
            params               += [thisparam]
            
            if length==None:
                varSizeParamLen   = len(fields[paramname])
        
        # create (dummy) buffer to hold reply
        reply                     = create_string_buffer(self.MAX_FRAME_LENGTH)
        
        # format params
        cparams                   = []
        cparams                  += [p['C'] for p in params]
        if varSizeParamLen!=None:
            cparams              += [varSizeParamLen]
        cparams                  += [byref(reply)]
        
        # log
        log.debug('C parameters: {0}'.format(pp.pformat(params)))
        
        # issue request
        rcApi     = getattr(self.smipmg,'dn_ipmg_{0}'.format(apiCommand['name']))(*cparams)
        assert rcApi == DN_ERR_NONE
        
        #===== wait for response
        
        if self.waitForResponse.wait(self.RESPONSE_TIMEOUT):
            self.waitForResponse.clear()
        else:
            self.smipmg.dn_ipmg_cancelTx()
            raise ConnectionError("timeout response")
        
        #===== convert C response to Python dictionnary
        
        # create returned fields on the fly
        theseFields = []
        for [name,type,length,validName] in apiCommand['response']['FIELDS']:
            theseFields += [(name, _getResponseType(type,length))]
        
        # Structure to hold the response
        class reply_t(Structure):
            _fields_    = theseFields
        
        # cast response to that structute
        notif           = cast(reply, POINTER(reply_t))
        
        # convert structure to dictionnary
        returnVal = {}
        for (f,_) in theseFields:
            rawVal = getattr(notif.contents,f)
            if  isinstance(rawVal,int):
                returnVal[f] = rawVal
            elif isinstance(rawVal,long):
                returnVal[f] = int(rawVal)
            else:
                returnVal[f] = [int(b) for b in rawVal]
        
        return returnVal
    
    def disconnect(self):
        
        # close serial port
        self.pyserialHandler.close()
        
        # disconnect the parent class
        ApiConnector.disconnect(self, reason="")
    
    #======================== private =========================================
    
    #========== C library interaction
    
    # UART interaction
    
    def _uart_txByte(self,b):
        '''
        \brief Called by C library when wants to send a byte over the serial
            port.
        '''
        log.debug('TX serial byte 0x{0:02x}'.format(b))
        
        self.pyserialHandler.write(chr(b))
        
        return 0
    
    # status notification
    
    def _ipmg_status(self,newStatus):
        '''
        \brief Called by C library when connection status is changed.
        '''
        log.debug('_ipmg_status newStatus={0}'.format(newStatus))
        
        if newStatus==DN_SERIAL_ST_CONNECTED:
            self.waitForConnected.set()
        
        return 0
    
    # data notifications
    
    def _ipmg_reply(self,cmdId):
        '''
        \brief Called by C library when a reply is received.
        '''
        log.debug('_ipmg_reply cmdId={0} (0x{0:02x})'.format(cmdId))
        
        self.waitForResponse.set()
        
        return 0
    
    def _ipmg_notif(self,cmdId,subcmdId):
        '''
        \brief Called by C library when a notification is received.
        '''
        log.debug(
            '_ipmg_notif cmdId={0} subcmdId={1} notifBuf={2}'.format(
                cmdId,
                subcmdId,
                FormatUtils.formatBuffer([ord(b) for b in self.notifBuf]),
            )
        )
        
        #===== cast notification buffer to correct type
        
        # find apiNotif
        apiNotifFields  = []
        nameArray       = ['notification']
        for notif in IpMgrDefinition.IpMgrDefinition.subCommandsNotification:
            if notif['id']==cmdId:
                nameArray += [notif['name']]
                for f in notif['response']['FIELDS']:
                    if f[0] not in [SUBID1,SUBID2]:
                        apiNotifFields += [f]
                if 'subCommands' in notif:
                    for subnotif in notif['subCommands']:
                        if subnotif['id']==subcmdId:
                            nameArray += [subnotif['name']]
                            for f in subnotif['response']['FIELDS']:
                                apiNotifFields += [f]
        
        # create returned fields on the fly
        theseFields = []
        lengthField = None
        for [name,type,length,_] in apiNotifFields:
            if length==None:
                theseFields += [('{0}Len'.format(name), c_uint8)]
                theseFields += [(name,                  c_uint8*self.MAX_FRAME_LENGTH)]
                lengthField  = name
            else:
                theseFields += [(name, _getResponseType(type,length))]
        
        # Structure to hold the response
        class reply_t(Structure):
            _fields_    = theseFields
        
        # cast response to that structute
        c_notif         = cast(self.notifBuf, POINTER(reply_t))
        
        #===== convert C notification to Python dictionnary
        
        # convert structure to dictionnary
        py_notif = {}
        for (f,_) in theseFields:
            rawVal = getattr(c_notif.contents,f)
            if  isinstance(rawVal,int):
                py_notif[f] = rawVal
            elif isinstance(rawVal,long):
                py_notif[f] = int(rawVal)
            else:
                py_notif[f] = [int(b) for b in rawVal]
        
        # trim lengthField
        if lengthField!=None:
            py_notif[lengthField] = py_notif[lengthField][:py_notif['{0}Len'.format(lengthField)]]
            del py_notif['{0}Len'.format(lengthField)]
        
        #===== put received packet in notification buffer
        
        self.putNotification((nameArray, py_notif))



from .telemetry_health_view import TelemetryHealthView
from .device_health_view import DeviceHealthView
from .configuration_edit import EditConfigurationView
from .asset_condition_view import AssetConditionsView
from .analytics_view import AnalyticsView
from .commit_view import CommitView
from .training_view import TrainingView
from .management_models_view import ManagementModelsView



from pydantic import BaseModel
from datetime import datetime
from typing import Dict


class AnnotationLabel(BaseModel):
    """This annotation label is the class that defines the contract
    of annotation labels for signals"""
    label_value: str
    parameters: Dict


class AnnotationLabelPayload(BaseModel):
    """This class gives all the context for an annotation of a signal
    in the persistence tables.
    """
    tenant_id: str
    edge_logical_id: str
    observation_timestamp: int
    label: AnnotationLabel
    version: str
    dataset_name: str



from txp.devices.tasks_scheduler import TasksScheduler, ScheduleTask
from txp.common.configuration import *
from txp.common.edge.common import MachineMetadata
from txp.devices.window_policy_manager import WindowPolicyManager
from txp.devices.telemetry_sender import TelemetrySender
from txp.devices.devices_manager import DevicesManager
from txp.common.config import settings
from functools import reduce
from typing import Optional
from queue import SimpleQueue
import time
import threading
import logging
import copy

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


class JobTasksManager:
    """The JobTasksManager is responsibly for the logic related to sampling
    tasks and telemetry sending in a job.

    """

    def __init__(
        self,
        gateway_conf: GatewayConfig,
        telemetry_sender: TelemetrySender,
        devices_manager: DevicesManager,
        configuration_id: Optional[str],
        machines: Optional[Dict[str, MachineMetadata]],
        job_task_finished_queue: SimpleQueue
    ):
        self._gateway_conf: GatewayConfig = gateway_conf
        self._configuration_id: str = configuration_id
        self._telemetry_sender: TelemetrySender = telemetry_sender
        self._devices_manager: DevicesManager = devices_manager
        self._machines_dict: Optional[Dict[str, MachineMetadata]] = machines
        self._cancel_event = threading.Event()

        self._window_managers: Dict[str, WindowPolicyManager] = {}
        self._scheduler: TasksScheduler = None
        self._job_task_finished_queue: SimpleQueue = job_task_finished_queue

    def start(self):
        self._scheduler = TasksScheduler("JobsTaskManager")
        self._scheduler.schedule_daily_callback(self._daily_schedule_plan_callback, "00:00:10")
        self._daily_schedule_plan_callback()
        self._scheduler.start()

    def stop(self):
        # Todo: ensure that all active WindowPolicyManager's are stopped
        self._cancel_event.set()
        if self._scheduler.is_running_tasks():
            self._scheduler.stop()
            while self._scheduler.is_running_tasks():
                time.sleep(1)

        if self._telemetry_sender.is_running():
            self._telemetry_sender.stop()
            while self._telemetry_sender.is_running():
                time.sleep(1)

        self._scheduler = None

    def _daily_schedule_plan_callback(self) -> None:
        """This is a callback intended to be called everyday to check which tasks
        require to be schedule for the current day.
        """
        sampling_tasks = self._get_sampling_tasks()
        telemetry_tasks = self._get_telemetry_tasks()

        for task_beg, task_end in sampling_tasks:
            if self._scheduler.is_schedule_required(task_beg):
                self._scheduler.schedule_task_once(task_beg)
                self._scheduler.schedule_task_once(task_end)

            elif self._scheduler.is_task_currently_active(task_beg):
                immediate_task = copy.copy(task_beg)
                immediate_task.start_time = datetime.datetime.now().time().replace(
                    microsecond=0, second=0
                )
                self._scheduler.schedule_task_once(
                    immediate_task,
                    daily=False,
                    log_info=True)
                self._scheduler.schedule_task_once(
                    task_end
                )

        for task in telemetry_tasks:
            if self._scheduler.is_schedule_required(task):
                self._scheduler.schedule_task_once(task)

            # Ad Hoc logic to check if Telemetry task already started
            elif self._is_telemetry_running(task):
                immediate_task = copy.copy(task)
                immediate_task.start_time = datetime.datetime.now().time().replace(
                    microsecond=0, second=0
                )
                now = datetime.datetime.now()
                leftover_time = abs(now - datetime.datetime.combine(
                    task.end_date, task.end_time
                )).total_seconds()

                task.callback_args = (task.callback_args[1], leftover_time)

                self._scheduler.schedule_task_once(
                    immediate_task,
                    daily=False,
                    log_info=True)

    def _is_telemetry_running(self, task: ScheduleTask) -> bool:
        """Returns True if the provided telemetry Task information should be currently
        running"""
        current_date = datetime.datetime.now().date()
        valid_date = (
                (task.start_date == current_date)
                or (current_date < task.end_date)
                or (current_date == task.end_date)
        )
        now = datetime.datetime.now()
        if task.end_date > now.date():
            valid_hour = (
                    task.start_time < now.time()
            )
        else:
            valid_hour = (
                    task.start_time < now.time() < task.end_time
            )

        # Validate the week day
        current_week_day = datetime.datetime.now().weekday()
        valid_week_day = bool(task.active_week_days[current_week_day])

        return valid_week_day and valid_date and valid_hour

    def _get_sampling_tasks(
        self
    ) -> List[ScheduleTask]:
        scheduler_tasks: List[(ScheduleTask, ScheduleTask)] = []
        for task in self._gateway_conf.job.tasks:
            edges_logical_ids: List[str] = []

            for asset_id in task.machines:
                edges_logical_ids = edges_logical_ids + \
                                    self._machines_dict[asset_id].edges_ids

            # Add start of task callback to the scheduler
            thread_name = str(time.time())
            callback_args = (thread_name, self._configuration_id, copy.deepcopy(task),
                             edges_logical_ids, self._devices_manager)

            sampling_task_begin = ScheduleTask(
                    task.parameters.start_date,
                    task.parameters.end_date,
                    task.parameters.start_time,
                    task.parameters.end_time,
                    task.parameters.week_days_mask,
                    self._start_sampling_task_callback,
                    callback_args
                )

            # And end of task callback to the scheduler
            sampling_task_end = ScheduleTask(
                    task.parameters.start_date,
                    task.parameters.end_date,
                    task.parameters.end_time,
                    task.parameters.end_time,
                    task.parameters.week_days_mask,
                    self._finish_sampling_task_callback,
                    (thread_name,),
                )

            scheduler_tasks.append((sampling_task_begin, sampling_task_end))

        return scheduler_tasks

    def _get_telemetry_tasks(self) -> List[ScheduleTask]:
        """Returns a 24hrs telemetry schedule plan for the provided gateway job configuration.

        Note: The user of the method should provide the callbacks for the returned entities.

        Returns:
            List of Telemetry Tasks
        """
        # Compute relative data
        current_date = datetime.datetime.now().date()
        job_start_time = self._gateway_conf.job.parameters.start_time
        job_end_time = self._gateway_conf.job.parameters.end_time
        job_sampling_window = self._gateway_conf.job.parameters.sampling_window

        if job_end_time <= job_start_time:  # The Job end in next day
            job_end_date = current_date + datetime.timedelta(days=1)
        else:
            job_end_date = current_date

        job_start_datetime = datetime.datetime.combine(current_date, job_start_time)
        job_end_datetime = datetime.datetime.combine(job_end_date, job_end_time)

        # Compute Telemetry Tasks
        total_job_secs = (job_end_datetime - job_start_datetime).total_seconds()
        num_sampling_windows = int(
            total_job_secs / job_sampling_window.observation_time
        )  # 24 hrs to seconds
        telemetry_tasks: List[ScheduleTask] = []

        logical_ids: List[str] = reduce(
            list.__add__,
            list(
                map(
                    lambda machine_metadata: machine_metadata.edges_ids,
                    self._machines_dict.values(),
                )
            ),
        )

        for i in range(0, num_sampling_windows):
            telemetry_start_time = job_start_datetime + datetime.timedelta(
                seconds=(
                    job_sampling_window.observation_time
                    + (job_sampling_window.observation_time * i)
                    - job_sampling_window.dead_time
                )
            )
            telemetry_end_time = telemetry_start_time + datetime.timedelta(
                seconds=job_sampling_window.dead_time
            )
            secs_timeout = (telemetry_end_time - telemetry_start_time).total_seconds()
            task = ScheduleTask(
                telemetry_start_time.date(),
                telemetry_end_time.date(),
                telemetry_start_time.time(),
                telemetry_end_time.time(),
                self._gateway_conf.job.parameters.week_days_mask,
                self._telemetry_sender.start,
                (self._gateway_conf, logical_ids, secs_timeout),
            )
            telemetry_tasks.append(task)

        log.info(
            f"[{self.__class__.__name__}]: The provided Job is active for current "
            f"day {current_date.strftime('%A')}. Job information: \n\t"
            f"Date: {current_date}. Day: {current_date.strftime('%A')}.\n\t"
            f"Total Telemetry Tasks computed: {len(telemetry_tasks)}"
        )

        return telemetry_tasks

    def _start_sampling_task_callback(
            self, window_manager_id: str, configuration_id: str, task: JobTask, logical_ids: List[str],
            devices_manager: DevicesManager
    ):
        window_manager: WindowPolicyManager = WindowPolicyManager(
            window_manager_id, configuration_id, self._gateway_conf.tenant_id, task, logical_ids, devices_manager,
            self._cancel_event
        )
        window_manager.start()
        self._window_managers[window_manager_id] = window_manager
        log.info(f"{self.__class__.__name__} starting a new " f"WindowPolicyManager")

    def _finish_sampling_task_callback(self, window_manager_id: str):
        window_manager = self._window_managers[window_manager_id]
        if window_manager.is_alive():
            log.warning(
                "Job Task hasn't finished yet, it's delayed. Nothing "
                "to do in current implementation"
            )
        else:
            log.info("Job Task finished successfully.")

        self._job_task_finished_queue.put(window_manager_id)

        self._window_managers.pop(window_manager_id)
